import asyncio
import time
import uuid
import warnings
from typing import Callable, Dict, List, Union

from ..client.api.drives.create_drive import asyncio as create_drive
from ..client.api.drives.create_drive import sync as create_drive_sync
from ..client.api.drives.delete_drive import asyncio as delete_drive
from ..client.api.drives.delete_drive import sync as delete_drive_sync
from ..client.api.drives.get_drive import asyncio as get_drive
from ..client.api.drives.get_drive import sync as get_drive_sync
from ..client.api.drives.list_drives import asyncio as list_drives
from ..client.api.drives.list_drives import sync as list_drives_sync
from ..client.api.drives.update_drive import asyncio as update_drive
from ..client.api.drives.update_drive import sync as update_drive_sync
from ..client.client import client
from ..client.errors import UnexpectedStatus
from ..client.models import Drive, DriveSpec, Metadata
from ..client.models.error import Error
from ..client.types import UNSET
from ..common.settings import settings


class DriveAPIError(Exception):
    """Exception raised when drive API returns an error."""

    def __init__(self, message: str, status_code: int | None = None, code: str | None = None):
        super().__init__(message)
        self.status_code = status_code
        self.code = code


class _AsyncDeleteDescriptor:
    """Descriptor that provides both class-level and instance-level delete functionality."""

    def __init__(self, delete_func: Callable):
        self._delete_func = delete_func

    def __get__(self, instance, owner):
        if instance is None:
            # Called on the class: DriveInstance.delete("name")
            return self._delete_func
        else:
            # Called on an instance: instance.delete()
            async def instance_delete() -> Drive:
                return await self._delete_func(instance.metadata.name or "")

            return instance_delete


class _SyncDeleteDescriptor:
    """Descriptor that provides both class-level and instance-level delete functionality (sync)."""

    def __init__(self, delete_func: Callable):
        self._delete_func = delete_func

    def __get__(self, instance, owner):
        if instance is None:
            # Called on the class: SyncDriveInstance.delete("name")
            return self._delete_func
        else:
            # Called on an instance: instance.delete()
            def instance_delete() -> Drive:
                return self._delete_func(instance.metadata.name or "")

            return instance_delete


class _AsyncUpdateDescriptor:
    """Descriptor that provides both class-level and instance-level update functionality."""

    def __init__(self, update_func: Callable):
        self._update_func = update_func

    def __get__(self, instance, owner):
        if instance is None:
            # Called on the class: DriveInstance.update("name", updates)
            return self._update_func
        else:
            # Called on an instance: instance.update(updates)
            async def instance_update(
                updates: Union["DriveCreateConfiguration", Drive, Dict[str, any]],
            ) -> "DriveInstance":
                return await self._update_func(instance.metadata.name or "", updates)

            return instance_update


class _SyncUpdateDescriptor:
    """Descriptor that provides both class-level and instance-level update functionality (sync)."""

    def __init__(self, update_func: Callable):
        self._update_func = update_func

    def __get__(self, instance, owner):
        if instance is None:
            # Called on the class: SyncDriveInstance.update("name", updates)
            return self._update_func
        else:
            # Called on an instance: instance.update(updates)
            def instance_update(
                updates: Union["DriveCreateConfiguration", Drive, Dict[str, any]],
            ) -> "SyncDriveInstance":
                return self._update_func(instance.metadata.name or "", updates)

            return instance_update


class DriveCreateConfiguration:
    """Simplified configuration for creating drives with default values."""

    def __init__(
        self,
        name: str | None = None,
        display_name: str | None = None,
        labels: Dict[str, str] | None = None,
        size: int | None = None,  # Size in GB
        region: str | None = None,  # Region
    ):
        self.name = name
        self.display_name = display_name
        self.labels = labels
        self.size = size
        self.region = region

    @classmethod
    def from_dict(cls, data: Dict[str, any]) -> "DriveCreateConfiguration":
        return cls(
            name=data.get("name"),
            display_name=data.get("display_name"),
            labels=data.get("labels"),
            size=data.get("size"),
            region=data.get("region"),
        )


class DriveInstance:
    delete: "_AsyncDeleteDescriptor"

    def __init__(self, drive: Drive):
        self.drive = drive

    @property
    def metadata(self):
        return self.drive.metadata

    @property
    def spec(self):
        return self.drive.spec

    @property
    def status(self):
        return self.drive.status

    @property
    def name(self):
        return self.drive.metadata.name if self.drive.metadata else None

    @property
    def display_name(self):
        return self.drive.metadata.display_name if self.drive.metadata else None

    @property
    def size(self):
        return self.drive.spec.size if self.drive.spec else None

    @property
    def region(self):
        return self.drive.spec.region if self.drive.spec else None

    @classmethod
    async def create(
        cls, config: Union[DriveCreateConfiguration, Drive, Dict[str, any]]
    ) -> "DriveInstance":
        # Generate default values
        default_name = f"drive-{uuid.uuid4().hex[:8]}"
        default_size = 10  # 10GB

        # Handle different configuration types
        if isinstance(config, Drive):
            drive = config
        elif isinstance(config, DriveCreateConfiguration):
            drive = Drive(
                metadata=Metadata(
                    name=config.name or default_name,
                    display_name=config.display_name or config.name or default_name,
                    labels=config.labels,
                ),
                spec=DriveSpec(
                    size=config.size or default_size,
                    region=config.region or settings.region or UNSET,
                ),
            )
        elif isinstance(config, dict):
            drive_config = DriveCreateConfiguration.from_dict(config)
            drive = Drive(
                metadata=Metadata(
                    name=drive_config.name or default_name,
                    display_name=drive_config.display_name or drive_config.name or default_name,
                    labels=drive_config.labels,
                ),
                spec=DriveSpec(
                    size=drive_config.size or default_size,
                    region=drive_config.region or settings.region or UNSET,
                ),
            )
        else:
            raise ValueError(
                f"Invalid config type: {type(config)}. Expected DriveCreateConfiguration, Drive, or dict."
            )

        # Ensure required fields have defaults
        if not drive.metadata:
            drive.metadata = Metadata(name=default_name)
        if not drive.metadata.name:
            drive.metadata.name = default_name
        if not drive.spec:
            drive.spec = DriveSpec(size=default_size)
        if not drive.spec.size:
            drive.spec.size = default_size

        # Warn if region is not set
        if not drive.spec.region or drive.spec.region is UNSET:
            warnings.warn(
                "DriveInstance.create: 'region' is not set. In a future version, 'region' will be a required parameter. "
                "Please specify a region (e.g. 'us-pdx-1', 'eu-lon-1', 'us-was-1') in the drive configuration or set the BL_REGION environment variable.",
                FutureWarning,
                stacklevel=2,
            )

        response = await create_drive(client=client, body=drive)
        if isinstance(response, Error):
            status_code = int(response.code) if response.code is not UNSET else None
            message = response.message if response.message is not UNSET else response.error
            raise DriveAPIError(message, status_code=status_code, code=response.error)
        return cls(response)

    @classmethod
    async def get(cls, drive_name: str) -> "DriveInstance":
        response = await get_drive(drive_name=drive_name, client=client)
        if response is None:
            raise DriveAPIError(
                f"Drive '{drive_name}' not found", status_code=404, code="NOT_FOUND"
            )
        if isinstance(response, Error):
            status_code = int(response.code) if response.code is not UNSET else None
            message = response.message if response.message is not UNSET else response.error
            raise DriveAPIError(message, status_code=status_code, code=response.error)
        return cls(response)

    @classmethod
    async def list(cls) -> list["DriveInstance"]:
        response = await list_drives(client=client)
        return [cls(drive) for drive in response or []]

    @classmethod
    async def create_if_not_exists(
        cls, config: Union[DriveCreateConfiguration, Drive, Dict[str, any]]
    ) -> "DriveInstance":
        """Create a drive if it doesn't exist, otherwise return existing."""
        try:
            return await cls.create(config)
        except (DriveAPIError, UnexpectedStatus) as e:
            is_conflict = False
            if isinstance(e, DriveAPIError):
                is_conflict = e.status_code == 409 or e.code in ["409", "DRIVE_ALREADY_EXISTS"]
            elif isinstance(e, UnexpectedStatus):
                is_conflict = e.status_code == 409

            if is_conflict:
                if isinstance(config, DriveCreateConfiguration):
                    name = config.name
                elif isinstance(config, dict):
                    name = config.get("name")
                elif isinstance(config, Drive):
                    name = config.metadata.name if config.metadata else None
                else:
                    name = None

                if not name:
                    raise ValueError("Drive name is required")

                return await cls.get(name)
            raise


class SyncDriveInstance:
    delete: "_SyncDeleteDescriptor"

    """Synchronous drive instance for managing persistent storage."""

    def __init__(self, drive: Drive):
        self.drive = drive

    @property
    def metadata(self):
        return self.drive.metadata

    @property
    def spec(self):
        return self.drive.spec

    @property
    def status(self):
        return self.drive.status

    @property
    def name(self):
        return self.drive.metadata.name if self.drive.metadata else None

    @property
    def display_name(self):
        return self.drive.metadata.display_name if self.drive.metadata else None

    @property
    def size(self):
        return self.drive.spec.size if self.drive.spec else None

    @property
    def region(self):
        return self.drive.spec.region if self.drive.spec else None

    @classmethod
    def create(
        cls, config: Union[DriveCreateConfiguration, Drive, Dict[str, any]]
    ) -> "SyncDriveInstance":
        """Create a new drive synchronously."""
        # Generate default values
        default_name = f"drive-{uuid.uuid4().hex[:8]}"
        default_size = 10  # 10GB

        # Handle different configuration types
        if isinstance(config, Drive):
            drive = config
        elif isinstance(config, DriveCreateConfiguration):
            drive = Drive(
                metadata=Metadata(
                    name=config.name or default_name,
                    display_name=config.display_name or config.name or default_name,
                    labels=config.labels,
                ),
                spec=DriveSpec(
                    size=config.size or default_size,
                    region=config.region or settings.region or UNSET,
                ),
            )
        elif isinstance(config, dict):
            drive_config = DriveCreateConfiguration.from_dict(config)
            drive = Drive(
                metadata=Metadata(
                    name=drive_config.name or default_name,
                    display_name=drive_config.display_name or drive_config.name or default_name,
                    labels=drive_config.labels,
                ),
                spec=DriveSpec(
                    size=drive_config.size or default_size,
                    region=drive_config.region or settings.region or UNSET,
                ),
            )
        else:
            raise ValueError(
                f"Invalid config type: {type(config)}. Expected DriveCreateConfiguration, Drive, or dict."
            )

        # Ensure required fields have defaults
        if not drive.metadata:
            drive.metadata = Metadata(name=default_name)
        if not drive.metadata.name:
            drive.metadata.name = default_name
        if not drive.spec:
            drive.spec = DriveSpec(size=default_size)
        if not drive.spec.size:
            drive.spec.size = default_size

        # Warn if region is not set
        if not drive.spec.region or drive.spec.region is UNSET:
            warnings.warn(
                "SyncDriveInstance.create: 'region' is not set. In a future version, 'region' will be a required parameter. "
                "Please specify a region (e.g. 'us-pdx-1', 'eu-lon-1', 'us-was-1') in the drive configuration or set the BL_REGION environment variable.",
                FutureWarning,
                stacklevel=2,
            )

        response = create_drive_sync(client=client, body=drive)
        if isinstance(response, Error):
            status_code = int(response.code) if response.code is not UNSET else None
            message = response.message if response.message is not UNSET else response.error
            raise DriveAPIError(message, status_code=status_code, code=response.error)
        return cls(response)

    @classmethod
    def get(cls, drive_name: str) -> "SyncDriveInstance":
        """Get a drive by name synchronously."""
        response = get_drive_sync(drive_name=drive_name, client=client)
        if response is None:
            raise DriveAPIError(
                f"Drive '{drive_name}' not found", status_code=404, code="NOT_FOUND"
            )
        if isinstance(response, Error):
            status_code = int(response.code) if response.code is not UNSET else None
            message = response.message if response.message is not UNSET else response.error
            raise DriveAPIError(message, status_code=status_code, code=response.error)
        return cls(response)

    @classmethod
    def list(cls) -> List["SyncDriveInstance"]:
        """List all drives synchronously."""
        response = list_drives_sync(client=client)
        return [cls(drive) for drive in response or []]

    @classmethod
    def create_if_not_exists(
        cls, config: Union[DriveCreateConfiguration, Drive, Dict[str, any]]
    ) -> "SyncDriveInstance":
        """Create a drive if it doesn't exist, otherwise return existing."""
        try:
            return cls.create(config)
        except (DriveAPIError, UnexpectedStatus) as e:
            is_conflict = False
            if isinstance(e, DriveAPIError):
                is_conflict = e.status_code == 409 or e.code in ["409", "DRIVE_ALREADY_EXISTS"]
            elif isinstance(e, UnexpectedStatus):
                is_conflict = e.status_code == 409

            if is_conflict:
                if isinstance(config, DriveCreateConfiguration):
                    name = config.name
                elif isinstance(config, dict):
                    name = config.get("name")
                elif isinstance(config, Drive):
                    name = config.metadata.name if config.metadata else None
                else:
                    name = None

                if not name:
                    raise ValueError("Drive name is required")

                return cls.get(name)
            raise


async def _delete_drive_by_name(drive_name: str) -> Drive:
    """Delete a drive by name (async)."""
    response = await delete_drive(drive_name=drive_name, client=client)
    return response


def _delete_drive_by_name_sync(drive_name: str) -> Drive:
    """Delete a drive by name (sync)."""
    response = delete_drive_sync(drive_name=drive_name, client=client)
    return response


async def _update_drive_by_name(
    drive_name: str, updates: Union[DriveCreateConfiguration, Drive, Dict[str, any]]
) -> "DriveInstance":
    """Update a drive by name (async)."""
    # Get the current drive
    drive_instance = await DriveInstance.get(drive_name)
    current_drive = drive_instance.drive

    # Build the update body
    if isinstance(updates, Drive):
        new_metadata = updates.metadata
        new_spec = updates.spec
    elif isinstance(updates, DriveCreateConfiguration):
        new_metadata = Metadata(
            name=current_drive.metadata.name if current_drive.metadata else drive_name,
            display_name=updates.display_name,
            labels=updates.labels,
        )
        new_spec = DriveSpec(
            size=updates.size,
            region=updates.region,
        )
    elif isinstance(updates, dict):
        config = DriveCreateConfiguration.from_dict(updates)
        new_metadata = Metadata(
            name=current_drive.metadata.name if current_drive.metadata else drive_name,
            display_name=config.display_name,
            labels=config.labels,
        )
        new_spec = DriveSpec(
            size=config.size,
            region=config.region,
        )
    else:
        raise ValueError(
            f"Invalid updates type: {type(updates)}. Expected DriveCreateConfiguration, Drive, or dict."
        )

    # Merge current values with updates
    merged_metadata = Metadata(
        name=current_drive.metadata.name if current_drive.metadata else drive_name,
        display_name=new_metadata.display_name
        if new_metadata and new_metadata.display_name
        else (current_drive.metadata.display_name if current_drive.metadata else None),
        labels=new_metadata.labels
        if new_metadata and new_metadata.labels
        else (current_drive.metadata.labels if current_drive.metadata else None),
    )

    merged_spec = DriveSpec(
        size=new_spec.size
        if new_spec and new_spec.size
        else (current_drive.spec.size if current_drive.spec else None),
        region=new_spec.region
        if new_spec and new_spec.region
        else (current_drive.spec.region if current_drive.spec else None),
    )

    body = Drive(
        metadata=merged_metadata,
        spec=merged_spec,
    )

    response = await update_drive(drive_name=drive_name, client=client, body=body)
    if isinstance(response, Error):
        status_code = int(response.code) if response.code is not UNSET else None
        message = response.message if response.message is not UNSET else response.error
        raise DriveAPIError(message, status_code=status_code, code=response.error)
    # This is for safe update
    await asyncio.sleep(0.5)
    return DriveInstance(response)


def _update_drive_by_name_sync(
    drive_name: str, updates: Union[DriveCreateConfiguration, Drive, Dict[str, any]]
) -> "SyncDriveInstance":
    """Update a drive by name (sync)."""
    # Get the current drive
    drive_instance = SyncDriveInstance.get(drive_name)
    current_drive = drive_instance.drive

    # Build the update body
    if isinstance(updates, Drive):
        new_metadata = updates.metadata
        new_spec = updates.spec
    elif isinstance(updates, DriveCreateConfiguration):
        new_metadata = Metadata(
            name=current_drive.metadata.name if current_drive.metadata else drive_name,
            display_name=updates.display_name,
            labels=updates.labels,
        )
        new_spec = DriveSpec(
            size=updates.size,
            region=updates.region,
        )
    elif isinstance(updates, dict):
        config = DriveCreateConfiguration.from_dict(updates)
        new_metadata = Metadata(
            name=current_drive.metadata.name if current_drive.metadata else drive_name,
            display_name=config.display_name,
            labels=config.labels,
        )
        new_spec = DriveSpec(
            size=config.size,
            region=config.region,
        )
    else:
        raise ValueError(
            f"Invalid updates type: {type(updates)}. Expected DriveCreateConfiguration, Drive, or dict."
        )

    # Merge current values with updates
    merged_metadata = Metadata(
        name=current_drive.metadata.name if current_drive.metadata else drive_name,
        display_name=new_metadata.display_name
        if new_metadata and new_metadata.display_name
        else (current_drive.metadata.display_name if current_drive.metadata else None),
        labels=new_metadata.labels
        if new_metadata and new_metadata.labels
        else (current_drive.metadata.labels if current_drive.metadata else None),
    )

    merged_spec = DriveSpec(
        size=new_spec.size
        if new_spec and new_spec.size
        else (current_drive.spec.size if current_drive.spec else None),
        region=new_spec.region
        if new_spec and new_spec.region
        else (current_drive.spec.region if current_drive.spec else None),
    )

    body = Drive(
        metadata=merged_metadata,
        spec=merged_spec,
    )

    response = update_drive_sync(drive_name=drive_name, client=client, body=body)
    if isinstance(response, Error):
        status_code = int(response.code) if response.code is not UNSET else None
        message = response.message if response.message is not UNSET else response.error
        raise DriveAPIError(message, status_code=status_code, code=response.error)
    # This is for safe update
    time.sleep(0.5)
    return SyncDriveInstance(response)


# Assign the delete descriptors to support both class-level and instance-level calls
DriveInstance.delete = _AsyncDeleteDescriptor(_delete_drive_by_name)
SyncDriveInstance.delete = _SyncDeleteDescriptor(_delete_drive_by_name_sync)

# Assign the update descriptors to support both class-level and instance-level calls
DriveInstance.update = _AsyncUpdateDescriptor(_update_drive_by_name)
SyncDriveInstance.update = _SyncUpdateDescriptor(_update_drive_by_name_sync)



"""Tests for WatchHandle, AsyncWatchHandle, StreamHandle, and AsyncStreamHandle classes."""

import pytest

from blaxel.core.sandbox import (
    AsyncStreamHandle,
    AsyncWatchHandle,
    StreamHandle,
    WatchHandle,
)


class TestWatchHandle:
    """Tests for WatchHandle (sync version)."""

    def test_close_calls_close_func(self):
        """Test that close() calls the close function."""
        close_called = False

        def close_func():
            nonlocal close_called
            close_called = True

        handle = WatchHandle(close_func)
        assert not close_called
        handle.close()
        assert close_called

    def test_close_only_calls_once(self):
        """Test that close() only calls the function once."""
        call_count = 0

        def close_func():
            nonlocal call_count
            call_count += 1

        handle = WatchHandle(close_func)
        handle.close()
        handle.close()
        handle.close()
        assert call_count == 1

    def test_closed_property(self):
        """Test the closed property."""
        handle = WatchHandle(lambda: None)
        assert not handle.closed
        handle.close()
        assert handle.closed

    def test_context_manager(self):
        """Test using WatchHandle as a context manager."""
        close_called = False

        def close_func():
            nonlocal close_called
            close_called = True

        with WatchHandle(close_func) as handle:
            assert not close_called
            assert not handle.closed

        assert close_called

    def test_context_manager_on_exception(self):
        """Test that context manager closes on exception."""
        close_called = False

        def close_func():
            nonlocal close_called
            close_called = True

        with pytest.raises(ValueError):
            with WatchHandle(close_func):
                raise ValueError("Test error")

        assert close_called

    def test_dict_like_access_backward_compatibility(self):
        """Test backward compatibility with dict-like access."""
        close_called = False

        def close_func():
            nonlocal close_called
            close_called = True

        handle = WatchHandle(close_func)
        # Should support handle["close"]() for backward compatibility
        handle["close"]()
        assert close_called

    def test_dict_like_access_invalid_key(self):
        """Test that invalid key raises KeyError."""
        handle = WatchHandle(lambda: None)
        with pytest.raises(KeyError):
            handle["invalid_key"]


class TestAsyncWatchHandle:
    """Tests for AsyncWatchHandle (async version)."""

    def test_close_calls_close_func(self):
        """Test that close() calls the close function."""
        close_called = False

        def close_func():
            nonlocal close_called
            close_called = True

        handle = AsyncWatchHandle(close_func)
        assert not close_called
        handle.close()
        assert close_called

    def test_close_only_calls_once(self):
        """Test that close() only calls the function once."""
        call_count = 0

        def close_func():
            nonlocal call_count
            call_count += 1

        handle = AsyncWatchHandle(close_func)
        handle.close()
        handle.close()
        handle.close()
        assert call_count == 1

    def test_closed_property(self):
        """Test the closed property."""
        handle = AsyncWatchHandle(lambda: None)
        assert not handle.closed
        handle.close()
        assert handle.closed

    @pytest.mark.asyncio
    async def test_async_context_manager(self):
        """Test using AsyncWatchHandle as an async context manager."""
        close_called = False

        def close_func():
            nonlocal close_called
            close_called = True

        async with AsyncWatchHandle(close_func) as handle:
            assert not close_called
            assert not handle.closed

        assert close_called

    @pytest.mark.asyncio
    async def test_async_context_manager_on_exception(self):
        """Test that async context manager closes on exception."""
        close_called = False

        def close_func():
            nonlocal close_called
            close_called = True

        with pytest.raises(ValueError):
            async with AsyncWatchHandle(close_func):
                raise ValueError("Test error")

        assert close_called

    def test_sync_context_manager(self):
        """Test that AsyncWatchHandle also supports sync context manager."""
        close_called = False

        def close_func():
            nonlocal close_called
            close_called = True

        with AsyncWatchHandle(close_func) as handle:
            assert not close_called
            assert not handle.closed

        assert close_called

    def test_dict_like_access_backward_compatibility(self):
        """Test backward compatibility with dict-like access."""
        close_called = False

        def close_func():
            nonlocal close_called
            close_called = True

        handle = AsyncWatchHandle(close_func)
        handle["close"]()
        assert close_called


class TestStreamHandle:
    """Tests for StreamHandle (sync version)."""

    def test_close_calls_close_func(self):
        """Test that close() calls the close function."""
        close_called = False

        def close_func():
            nonlocal close_called
            close_called = True

        handle = StreamHandle(close_func)
        assert not close_called
        handle.close()
        assert close_called

    def test_close_only_calls_once(self):
        """Test that close() only calls the function once."""
        call_count = 0

        def close_func():
            nonlocal call_count
            call_count += 1

        handle = StreamHandle(close_func)
        handle.close()
        handle.close()
        handle.close()
        assert call_count == 1

    def test_closed_property(self):
        """Test the closed property."""
        handle = StreamHandle(lambda: None)
        assert not handle.closed
        handle.close()
        assert handle.closed

    def test_context_manager(self):
        """Test using StreamHandle as a context manager."""
        close_called = False

        def close_func():
            nonlocal close_called
            close_called = True

        with StreamHandle(close_func) as handle:
            assert not close_called
            assert not handle.closed

        assert close_called

    def test_context_manager_on_exception(self):
        """Test that context manager closes on exception."""
        close_called = False

        def close_func():
            nonlocal close_called
            close_called = True

        with pytest.raises(ValueError):
            with StreamHandle(close_func):
                raise ValueError("Test error")

        assert close_called

    def test_dict_like_access_backward_compatibility(self):
        """Test backward compatibility with dict-like access."""
        close_called = False

        def close_func():
            nonlocal close_called
            close_called = True

        handle = StreamHandle(close_func)
        handle["close"]()
        assert close_called


class TestAsyncStreamHandle:
    """Tests for AsyncStreamHandle (async version)."""

    def test_close_calls_close_func(self):
        """Test that close() calls the close function."""
        close_called = False

        def close_func():
            nonlocal close_called
            close_called = True

        handle = AsyncStreamHandle(close_func)
        assert not close_called
        handle.close()
        assert close_called

    def test_close_only_calls_once(self):
        """Test that close() only calls the function once."""
        call_count = 0

        def close_func():
            nonlocal call_count
            call_count += 1

        handle = AsyncStreamHandle(close_func)
        handle.close()
        handle.close()
        handle.close()
        assert call_count == 1

    def test_closed_property(self):
        """Test the closed property."""
        handle = AsyncStreamHandle(lambda: None)
        assert not handle.closed
        handle.close()
        assert handle.closed

    @pytest.mark.asyncio
    async def test_async_context_manager(self):
        """Test using AsyncStreamHandle as an async context manager."""
        close_called = False

        def close_func():
            nonlocal close_called
            close_called = True

        async with AsyncStreamHandle(close_func) as handle:
            assert not close_called
            assert not handle.closed

        assert close_called

    @pytest.mark.asyncio
    async def test_async_context_manager_on_exception(self):
        """Test that async context manager closes on exception."""
        close_called = False

        def close_func():
            nonlocal close_called
            close_called = True

        with pytest.raises(ValueError):
            async with AsyncStreamHandle(close_func):
                raise ValueError("Test error")

        assert close_called

    def test_sync_context_manager(self):
        """Test that AsyncStreamHandle also supports sync context manager."""
        close_called = False

        def close_func():
            nonlocal close_called
            close_called = True

        with AsyncStreamHandle(close_func) as handle:
            assert not close_called
            assert not handle.closed

        assert close_called

    def test_dict_like_access_backward_compatibility(self):
        """Test backward compatibility with dict-like access."""
        close_called = False

        def close_func():
            nonlocal close_called
            close_called = True

        handle = AsyncStreamHandle(close_func)
        handle["close"]()
        assert close_called



import os
import platform
from pathlib import Path
from typing import Dict

import yaml

from ..authentication import BlaxelAuth, auth
from .logger import init_logger


def _get_os_arch() -> str:
    """Get OS and architecture information."""
    try:
        system = platform.system().lower()
        if system == "windows":
            os_name = "windows"
        elif system == "darwin":
            os_name = "darwin"
        elif system == "linux":
            os_name = "linux"
        else:
            os_name = system

        machine = platform.machine().lower()
        if machine in ["x86_64", "amd64"]:
            arch = "amd64"
        elif machine in ["aarch64", "arm64"]:
            arch = "arm64"
        elif machine in ["i386", "i686", "x86"]:
            arch = "386"
        else:
            arch = machine

        return f"{os_name}/{arch}"
    except Exception:
        return "unknown/unknown"


class Settings:
    auth: BlaxelAuth

    def __init__(self):
        init_logger(self.log_level)
        self.auth = auth(self.env, self.base_url)
        self._headers = None

    @property
    def env(self) -> str:
        """Get the environment."""
        return os.environ.get("BL_ENV", "prod")

    @property
    def log_level(self) -> str:
        """Get the log level."""
        return os.environ.get("LOG_LEVEL", "INFO").upper()

    @property
    def base_url(self) -> str:
        """Get the base URL for the API."""
        if self.env == "prod":
            return "https://api.blaxel.ai/v0"
        return "https://api.blaxel.dev/v0"

    @property
    def run_url(self) -> str:
        """Get the run URL."""
        if self.env == "prod":
            return "https://run.blaxel.ai"
        return "https://run.blaxel.dev"

    @property
    def sentry_dsn(self) -> str:
        """Get the Sentry DSN (injected at build time)."""
        import blaxel

        return blaxel.__sentry_dsn__

    @property
    def version(self) -> str:
        """Get the package version (injected at build time)."""
        import blaxel

        return blaxel.__version__ or "unknown"

    @property
    def commit(self) -> str:
        """Get the commit hash (injected at build time)."""
        import blaxel

        return blaxel.__commit__ or "unknown"

    @property
    def headers(self) -> Dict[str, str]:
        """Get the headers for API requests."""
        headers = self.auth.get_headers()
        os_arch = _get_os_arch()
        headers["User-Agent"] = f"blaxel/sdk/python/{self.version} ({os_arch}) blaxel/{self.commit}"
        return headers

    @property
    def name(self) -> str:
        """Get the name."""
        return os.environ.get("BL_NAME", "")

    @property
    def type(self) -> str:
        """Get the type."""
        return os.environ.get("BL_TYPE", "agent")

    @property
    def workspace(self) -> str:
        """Get the workspace."""
        return self.auth.workspace_name

    @property
    def run_internal_hostname(self) -> str:
        """Get the run internal hostname."""
        if self.generation == "":
            return ""
        return os.environ.get("BL_RUN_INTERNAL_HOST", "")

    @property
    def generation(self) -> str:
        """Get the generation."""
        return os.environ.get("BL_GENERATION", "")

    @property
    def bl_cloud(self) -> bool:
        """Is running on bl cloud."""
        return os.environ.get("BL_CLOUD", "") == "true"

    @property
    def run_internal_protocol(self) -> str:
        """Get the run internal protocol."""
        return os.environ.get("BL_RUN_INTERNAL_PROTOCOL", "https")

    @property
    def enable_opentelemetry(self) -> bool:
        """Get the enable opentelemetry."""
        return os.getenv("BL_ENABLE_OPENTELEMETRY", "false").lower() == "true"

    @property
    def tracking(self) -> bool:
        """
        Get the tracking setting.

        Priority:
        1. Environment variable DO_NOT_TRACK (true/1 to disable, false/0 to enable)
        2. config.yaml tracking field
        3. Default: true
        """
        env_value = os.environ.get("DO_NOT_TRACK")
        if env_value is not None:
            return env_value.lower() not in ("true", "1")

        try:
            home_dir = Path.home()
            config_path = home_dir / ".blaxel" / "config.yaml"
            with open(config_path, encoding="utf-8") as f:
                config = yaml.safe_load(f)
            if config and "tracking" in config:
                return bool(config["tracking"])
        except Exception:
            pass

        return True

    @property
    def region(self) -> str | None:
        """Get the region from BL_REGION environment variable."""
        return os.environ.get("BL_REGION")


settings = Settings()



import asyncio

import pytest

from blaxel.core import SandboxInstance
from blaxel.core.sandbox.types import SandboxUpdateMetadata
from tests.helpers import (
    default_image,
    default_labels,
    default_region,
    unique_name,
    wait_for_sandbox_deletion,
)

# =============================================================================
# Sandbox Create Tests
# =============================================================================


@pytest.mark.asyncio(loop_scope="class")
class TestSandboxCreate:
    """Test sandbox creation operations."""

    async def test_creates_sandbox_with_default_settings(self):
        """Test creating a sandbox with default settings."""
        sandbox = await SandboxInstance.create({"labels": default_labels})

        try:
            assert sandbox.metadata.name is not None
            assert sandbox.metadata.name.startswith("sandbox-")
        finally:
            await SandboxInstance.delete(sandbox.metadata.name)

    async def test_creates_sandbox_with_custom_name(self):
        """Test creating a sandbox with a custom name."""
        name = unique_name("custom")
        sandbox = await SandboxInstance.create({"name": name, "labels": default_labels})

        try:
            assert sandbox.metadata.name == name
        finally:
            await SandboxInstance.delete(name)

    async def test_creates_sandbox_with_specific_image(self):
        """Test creating a sandbox with a specific image."""
        name = unique_name("image-test")
        sandbox = await SandboxInstance.create(
            {
                "name": name,
                "image": default_image,
                "labels": default_labels,
            }
        )

        try:
            assert sandbox.metadata.name == name
        finally:
            await SandboxInstance.delete(name)

    async def test_creates_sandbox_with_memory_configuration(self):
        """Test creating a sandbox with memory configuration."""
        name = unique_name("memory-test")
        await SandboxInstance.create(
            {
                "name": name,
                "image": default_image,
                "memory": 2048,
                "labels": default_labels,
            }
        )

        try:
            retrieved = await SandboxInstance.get(name)
            assert retrieved.spec.runtime.memory == 2048
        finally:
            await SandboxInstance.delete(name)

    async def test_creates_sandbox_with_labels(self):
        """Test creating a sandbox with labels."""
        name = unique_name("labels-test")
        sandbox = await SandboxInstance.create(
            {
                "name": name,
                "labels": {**default_labels, "env": "test", "purpose": "integration"},
            }
        )

        try:
            assert sandbox.metadata.labels["env"] == "test"
            assert sandbox.metadata.labels["purpose"] == "integration"
        finally:
            await SandboxInstance.delete(name)

    async def test_creates_sandbox_with_ports(self):
        """Test creating a sandbox with ports."""
        name = unique_name("ports-test")
        await SandboxInstance.create(
            {
                "name": name,
                "image": default_image,
                "memory": 2048,
                "ports": [
                    {"name": "web", "target": 3000},
                    {"name": "api", "target": 8080, "protocol": "TCP"},
                ],
                "labels": default_labels,
            }
        )

        try:
            retrieved = await SandboxInstance.get(name)
            assert len(retrieved.spec.runtime.ports) == 2
        finally:
            await SandboxInstance.delete(name)

    async def test_creates_sandbox_with_environment_variables(self):
        """Test creating a sandbox with environment variables."""
        name = unique_name("envs-test")
        await SandboxInstance.create(
            {
                "name": name,
                "image": default_image,
                "envs": [
                    {"name": "NODE_ENV", "value": "test"},
                    {"name": "DEBUG", "value": "true"},
                ],
                "labels": default_labels,
            }
        )

        try:
            retrieved = await SandboxInstance.get(name)
            assert len(retrieved.spec.runtime.envs) == 2
        finally:
            await SandboxInstance.delete(name)

    async def test_creates_sandbox_with_region(self):
        """Test creating a sandbox with a region."""
        name = unique_name("region-test")
        await SandboxInstance.create(
            {
                "name": name,
                "image": default_image,
                "region": default_region,
                "labels": default_labels,
            }
        )

        try:
            retrieved = await SandboxInstance.get(name)
            assert retrieved.spec.region == default_region
        finally:
            await SandboxInstance.delete(name)


# =============================================================================
# Sandbox CreateIfNotExists Tests
# =============================================================================


@pytest.mark.asyncio(loop_scope="class")
class TestSandboxCreateIfNotExists:
    """Test createIfNotExists operations."""

    async def test_creates_new_sandbox_if_not_exists(self):
        """Test creating a new sandbox if it doesn't exist."""
        name = unique_name("cine")
        sandbox = await SandboxInstance.create_if_not_exists(
            {"name": name, "labels": default_labels}
        )

        try:
            assert sandbox.metadata.name == name
        finally:
            await SandboxInstance.delete(name)

    async def test_returns_existing_sandbox_if_exists(self):
        """Test returning existing sandbox if it already exists."""
        name = unique_name("cine-existing")

        # Create first
        first = await SandboxInstance.create({"name": name, "labels": default_labels})

        try:
            # create_if_not_exists should return the same sandbox
            second = await SandboxInstance.create_if_not_exists(
                {"name": name, "labels": default_labels}
            )
            assert second.metadata.name == first.metadata.name
        finally:
            await SandboxInstance.delete(name)

    async def test_handles_concurrent_create_if_not_exists_calls(self):
        """Test handling concurrent createIfNotExists calls."""
        name = unique_name("cine-race")
        concurrent_calls = 5

        async def create_task():
            try:
                sb = await SandboxInstance.create_if_not_exists(
                    {"name": name, "labels": default_labels}
                )
                return {"sandbox": sb, "error": None}
            except Exception as e:
                return {"sandbox": None, "error": e}

        results = await asyncio.gather(
            *[create_task() for _ in range(concurrent_calls)],
            return_exceptions=True,
        )

        try:
            successes = [r for r in results if isinstance(r, dict) and r.get("sandbox") is not None]
            unique_names = set(r["sandbox"].metadata.name for r in successes)

            assert len(unique_names) == 1
            assert len(successes) >= 2

            sandbox = await SandboxInstance.get(name)
            result = await sandbox.process.exec(
                {"command": "echo 'test'", "wait_for_completion": True}
            )
            assert result.logs == "test\n"
        finally:
            await SandboxInstance.delete(name)


# =============================================================================
# Sandbox Get Tests
# =============================================================================


@pytest.mark.asyncio(loop_scope="class")
class TestSandboxGet:
    """Test sandbox get operations."""

    async def test_retrieves_existing_sandbox(self):
        """Test retrieving an existing sandbox."""
        name = unique_name("get-test")
        await SandboxInstance.create({"name": name, "labels": default_labels})

        try:
            retrieved = await SandboxInstance.get(name)
            assert retrieved.metadata.name == name
        finally:
            await SandboxInstance.delete(name)

    async def test_throws_error_for_non_existent_sandbox(self):
        """Test that getting a non-existent sandbox raises an error."""
        with pytest.raises(Exception):
            await SandboxInstance.get("non-existent-sandbox-xyz")


# =============================================================================
# Sandbox List Tests
# =============================================================================


@pytest.mark.asyncio(loop_scope="class")
class TestSandboxList:
    """Test sandbox list operations."""

    async def test_lists_all_sandboxes(self):
        """Test listing all sandboxes."""
        name = unique_name("list-test")
        await SandboxInstance.create({"name": name, "labels": default_labels})

        try:
            sandboxes = await SandboxInstance.list()
            assert isinstance(sandboxes, list)

            found = next((s for s in sandboxes if s.metadata.name == name), None)
            assert found is not None
        finally:
            await SandboxInstance.delete(name)


# =============================================================================
# Sandbox Delete Tests
# =============================================================================


@pytest.mark.asyncio(loop_scope="class")
class TestSandboxDelete:
    """Test sandbox delete operations."""

    async def test_deletes_existing_sandbox(self):
        """Test deleting an existing sandbox."""
        name = unique_name("delete-test")
        await SandboxInstance.create({"name": name, "labels": default_labels})

        await SandboxInstance.delete(name)

        # Wait for deletion to fully complete
        deleted = await wait_for_sandbox_deletion(name)
        assert deleted is True

    async def test_can_delete_using_instance_method(self):
        """Test deleting using instance method."""
        name = unique_name("delete-instance")
        sandbox = await SandboxInstance.create({"name": name, "labels": default_labels})

        await sandbox.delete()

        # Wait for deletion to fully complete
        deleted = await wait_for_sandbox_deletion(name)
        assert deleted is True


# =============================================================================
# Sandbox UpdateMetadata Tests
# =============================================================================


@pytest.mark.asyncio(loop_scope="class")
class TestSandboxUpdateMetadata:
    """Test sandbox updateMetadata operations."""

    async def test_updates_sandbox_labels(self):
        """Test updating sandbox labels."""
        name = unique_name("update-meta")
        await SandboxInstance.create({"name": name, "labels": default_labels})

        try:
            updated = await SandboxInstance.update_metadata(
                name,
                SandboxUpdateMetadata(labels={**default_labels, "updated": "true"}),
            )
            assert updated.metadata.labels["updated"] == "true"
        finally:
            await SandboxInstance.delete(name)

    async def test_updates_sandbox_display_name(self):
        """Test updating sandbox displayName."""
        name = unique_name("update-display")
        await SandboxInstance.create({"name": name, "labels": default_labels})

        try:
            updated = await SandboxInstance.update_metadata(
                name,
                SandboxUpdateMetadata(display_name="My Test Sandbox"),
            )
            assert updated.metadata.display_name == "My Test Sandbox"
        finally:
            await SandboxInstance.delete(name)


# =============================================================================
# Sandbox Wait Tests
# =============================================================================


@pytest.mark.asyncio(loop_scope="class")
class TestSandboxWait:
    """Test sandbox wait operations."""

    async def test_waits_for_sandbox_to_be_ready(self):
        """Test waiting for sandbox to be ready."""
        name = unique_name("wait-test")
        sandbox = await SandboxInstance.create({"name": name, "labels": default_labels})

        try:
            # After wait, sandbox should be ready and we can run commands
            result = await sandbox.process.exec(
                {"command": "echo 'ready'", "wait_for_completion": True}
            )
            assert "ready" in result.logs
        finally:
            await SandboxInstance.delete(name)



"""Integration tests for Image build functionality."""

import uuid

import pytest
import pytest_asyncio

from blaxel.core import ImageInstance, SandboxInstance


@pytest.mark.asyncio(loop_scope="class")
class TestImage:
    """Test re-building the same image with an updated spec (update scenario)."""

    image_name: str
    sandbox: SandboxInstance

    @pytest_asyncio.fixture(autouse=True, scope="class", loop_scope="class")
    async def setup_rebuilt_image(self, request):
        image_name = f"image-test-{uuid.uuid4().hex[:8]}"
        request.cls.image_name = image_name

        # Build initial image (v1)
        image_v1 = (
            ImageInstance.from_registry("python:3.11-slim")
            .apt_install("curl", "git", "wget")
            .workdir("/app")
            .pip_install("requests", "httpx", "pydantic")
            .env(PYTHONUNBUFFERED="1", APP_NAME="blaxel-test")
            .copy(".", "/app")
            .expose(8080)
        )

        await image_v1.build(
            name=image_name,
            memory=4096,
            timeout=900.0,
            sandbox_version="latest",
        )

        # Rebuild with updated spec (v2)
        image_v2 = (
            ImageInstance.from_registry("python:3.11-alpine")
            .apk_add("curl", "git", "wget", "vim")
            .workdir("/app")
            .pip_install("requests", "httpx", "pydantic", "aiohttp")
            .env(
                PYTHONUNBUFFERED="1",
                APP_NAME="blaxel-test",
                DEBUG="true",
                LOG_LEVEL="debug",
                VERSION="2.0.0",
            )
            .label(
                maintainer="blaxel",
                version="2.0.0",
                description="Integration test image v2",
            )
            .run_commands(
                "python --version",
                "curl --version",
                "git --version",
                "vim --version | head -1",
            )
            .copy(".", "/app")
            .expose(8080)
        )

        await image_v2.build(
            name=image_name,
            memory=4096,
            timeout=900.0,
            sandbox_version="latest",
        )

        request.cls.sandbox = await SandboxInstance.create(
            {
                "image": image_name,
                "memory": 4096,
            }
        )

        yield

        await SandboxInstance.delete(self.sandbox.metadata.name)

    async def test_new_python_package_available(self):
        result = await self.sandbox.process.exec(
            {
                "command": "python -c \"import aiohttp; print('aiohttp OK')\"",
                "waitForCompletion": True,
            }
        )
        process_name = result.name if result.name else "unknown"
        logs = await self.sandbox.process.logs(process_name, "all")

        assert logs is not None
        assert "aiohttp OK" in logs

    async def test_new_apt_package_available(self):
        result = await self.sandbox.process.exec(
            {
                "command": "vim --version | head -1",
                "waitForCompletion": True,
            }
        )
        process_name = result.name if result.name else "unknown"
        logs = await self.sandbox.process.logs(process_name, "all")

        assert logs is not None
        assert "vim" in logs.lower()



from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.sandbox_lifecycle import SandboxLifecycle
    from ..models.sandbox_network import SandboxNetwork
    from ..models.sandbox_runtime import SandboxRuntime
    from ..models.volume_attachment import VolumeAttachment


T = TypeVar("T", bound="SandboxSpec")


@_attrs_define
class SandboxSpec:
    """Configuration for a sandbox including its image, memory, ports, region, and lifecycle policies

    Attributes:
        enabled (Union[Unset, bool]): When false, the sandbox is disabled and will not accept connections Default: True.
            Example: True.
        lifecycle (Union[Unset, SandboxLifecycle]): Lifecycle configuration controlling automatic sandbox deletion based
            on idle time, max age, or specific dates
        network (Union[Unset, SandboxNetwork]): Network configuration for a sandbox including egress IP binding. All
            three fields (vpcName, egressGatewayName, egressIpName) must be specified together to assign a dedicated IP.
        region (Union[Unset, str]): Region where the sandbox should be created (e.g. us-pdx-1, eu-lon-1). If not
            specified, defaults to the region closest to the user. Example: us-pdx-1.
        runtime (Union[Unset, SandboxRuntime]): Runtime configuration defining how the sandbox VM is provisioned and its
            resource limits
        volumes (Union[Unset, list['VolumeAttachment']]):
    """

    enabled: Union[Unset, bool] = True
    lifecycle: Union[Unset, "SandboxLifecycle"] = UNSET
    network: Union[Unset, "SandboxNetwork"] = UNSET
    region: Union[Unset, str] = UNSET
    runtime: Union[Unset, "SandboxRuntime"] = UNSET
    volumes: Union[Unset, list["VolumeAttachment"]] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        enabled = self.enabled

        lifecycle: Union[Unset, dict[str, Any]] = UNSET
        if (
            self.lifecycle
            and not isinstance(self.lifecycle, Unset)
            and not isinstance(self.lifecycle, dict)
        ):
            lifecycle = self.lifecycle.to_dict()
        elif self.lifecycle and isinstance(self.lifecycle, dict):
            lifecycle = self.lifecycle

        network: Union[Unset, dict[str, Any]] = UNSET
        if (
            self.network
            and not isinstance(self.network, Unset)
            and not isinstance(self.network, dict)
        ):
            network = self.network.to_dict()
        elif self.network and isinstance(self.network, dict):
            network = self.network

        region = self.region

        runtime: Union[Unset, dict[str, Any]] = UNSET
        if (
            self.runtime
            and not isinstance(self.runtime, Unset)
            and not isinstance(self.runtime, dict)
        ):
            runtime = self.runtime.to_dict()
        elif self.runtime and isinstance(self.runtime, dict):
            runtime = self.runtime

        volumes: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.volumes, Unset):
            volumes = []
            for componentsschemas_volume_attachments_item_data in self.volumes:
                if type(componentsschemas_volume_attachments_item_data) is dict:
                    componentsschemas_volume_attachments_item = (
                        componentsschemas_volume_attachments_item_data
                    )
                else:
                    componentsschemas_volume_attachments_item = (
                        componentsschemas_volume_attachments_item_data.to_dict()
                    )
                volumes.append(componentsschemas_volume_attachments_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if enabled is not UNSET:
            field_dict["enabled"] = enabled
        if lifecycle is not UNSET:
            field_dict["lifecycle"] = lifecycle
        if network is not UNSET:
            field_dict["network"] = network
        if region is not UNSET:
            field_dict["region"] = region
        if runtime is not UNSET:
            field_dict["runtime"] = runtime
        if volumes is not UNSET:
            field_dict["volumes"] = volumes

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T | None:
        from ..models.sandbox_lifecycle import SandboxLifecycle
        from ..models.sandbox_network import SandboxNetwork
        from ..models.sandbox_runtime import SandboxRuntime
        from ..models.volume_attachment import VolumeAttachment

        if not src_dict:
            return None
        d = src_dict.copy()
        enabled = d.pop("enabled", UNSET)

        _lifecycle = d.pop("lifecycle", UNSET)
        lifecycle: Union[Unset, SandboxLifecycle]
        if isinstance(_lifecycle, Unset):
            lifecycle = UNSET
        else:
            lifecycle = SandboxLifecycle.from_dict(_lifecycle)

        _network = d.pop("network", UNSET)
        network: Union[Unset, SandboxNetwork]
        if isinstance(_network, Unset):
            network = UNSET
        else:
            network = SandboxNetwork.from_dict(_network)

        region = d.pop("region", UNSET)

        _runtime = d.pop("runtime", UNSET)
        runtime: Union[Unset, SandboxRuntime]
        if isinstance(_runtime, Unset):
            runtime = UNSET
        else:
            runtime = SandboxRuntime.from_dict(_runtime)

        volumes = []
        _volumes = d.pop("volumes", UNSET)
        for componentsschemas_volume_attachments_item_data in _volumes or []:
            componentsschemas_volume_attachments_item = VolumeAttachment.from_dict(
                componentsschemas_volume_attachments_item_data
            )

            volumes.append(componentsschemas_volume_attachments_item)

        sandbox_spec = cls(
            enabled=enabled,
            lifecycle=lifecycle,
            network=network,
            region=region,
            runtime=runtime,
            volumes=volumes,
        )

        sandbox_spec.additional_properties = d
        return sandbox_spec

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""MCP Client Integration Tests.

Note: These tests require special authentication setup for raw MCP SDK.
The blTools wrapper in test_tools.py handles auth automatically.
"""

import pytest
import pytest_asyncio
from mcp import ClientSession
from mcp.client.streamable_http import streamablehttp_client

from blaxel.core import settings
from blaxel.core.sandbox import SandboxInstance
from tests.helpers import default_image, default_labels, unique_name


@pytest.mark.asyncio(loop_scope="class")
class TestMCPClientIntegration:
    """Test MCP client integration."""

    sandbox: SandboxInstance | None = None
    sandbox_name: str | None = None

    @pytest_asyncio.fixture(autouse=True, scope="class", loop_scope="class")
    async def setup_sandbox(self, request):
        """Set up a sandbox for the test class."""
        request.cls.sandbox_name = unique_name("mcp-test")
        request.cls.sandbox = await SandboxInstance.create(
            {
                "name": request.cls.sandbox_name,
                "image": default_image,
                "memory": 2048,
                "labels": default_labels,
            }
        )

        yield

        # Cleanup
        try:
            await request.cls.sandbox.delete()
        except Exception:
            pass

    async def test_streamable_http_transport(self):
        """Test Streamable HTTP Transport."""
        if not self.sandbox:
            pytest.skip("Sandbox not found")
        base_url = f"{self.sandbox.metadata.url}/mcp"

        async with streamablehttp_client(
            base_url,
            headers=settings.headers,
        ) as (read_stream, write_stream, _):
            async with ClientSession(read_stream, write_stream) as session:
                await session.initialize()

                # Verify connection worked
                assert session is not None

                response = await session.list_tools()

                assert response is not None



import time
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, List, Union

from ...client import errors
from ...client.api.compute.create_sandbox_preview import sync as create_sandbox_preview
from ...client.api.compute.create_sandbox_preview_token import (
    sync as create_sandbox_preview_token,
)
from ...client.api.compute.delete_sandbox_preview import sync as delete_sandbox_preview
from ...client.api.compute.delete_sandbox_preview_token import (
    sync as delete_sandbox_preview_token,
)
from ...client.api.compute.get_sandbox_preview import sync as get_sandbox_preview
from ...client.api.compute.get_sandbox_preview import (
    sync_detailed as get_sandbox_preview_detailed,
)
from ...client.api.compute.list_sandbox_preview_tokens import (
    sync as list_sandbox_preview_tokens,
)
from ...client.api.compute.list_sandbox_previews import sync as list_sandbox_previews
from ...client.client import client
from ...client.models import (
    Preview,
    PreviewMetadata,
    PreviewSpec,
    PreviewToken,
    PreviewTokenSpec,
    Sandbox,
)


@dataclass
class SyncSandboxPreviewToken:
    preview_token: PreviewToken

    @property
    def value(self) -> str:
        return (
            self.preview_token.spec.token
            if self.preview_token.spec and self.preview_token.spec.token
            else ""
        )

    @property
    def expires_at(self) -> datetime:
        return (
            datetime.fromisoformat(self.preview_token.spec.expires_at)
            if self.preview_token.spec and self.preview_token.spec.expires_at
            else datetime.now()
        )


class SyncSandboxPreviewTokens:
    def __init__(self, preview: Preview):
        self.preview = preview

    @property
    def preview_name(self) -> str:
        return (
            self.preview.metadata.name
            if self.preview.metadata and self.preview.metadata.name
            else ""
        )

    @property
    def resource_name(self) -> str:
        return (
            self.preview.metadata.resource_name
            if self.preview.metadata and self.preview.metadata.resource_name
            else ""
        )

    def create(self, expires_at: datetime):
        response = create_sandbox_preview_token(
            self.resource_name,
            self.preview_name,
            body=PreviewToken(
                spec=PreviewTokenSpec(
                    expires_at=to_utc_z(expires_at),
                )
            ),
            client=client,
        )
        if response:
            return SyncSandboxPreviewToken(response)
        raise errors.UnexpectedStatus(400, b"Failed to create preview token")

    def list(self):
        response = list_sandbox_preview_tokens(
            self.resource_name,
            self.preview_name,
            client=client,
        )
        if response:
            return [SyncSandboxPreviewToken(token) for token in response]
        raise errors.UnexpectedStatus(400, b"Failed to list preview tokens")

    def delete(self, token_name: str):
        response = delete_sandbox_preview_token(
            self.resource_name,
            self.preview_name,
            token_name,
            client=client,
        )
        if response:
            return response
        raise errors.UnexpectedStatus(400, b"Failed to delete preview token")


class SyncSandboxPreview:
    def __init__(self, preview: Preview):
        self.preview = preview
        self.tokens = SyncSandboxPreviewTokens(preview)

    @property
    def name(self) -> str:
        return (
            self.preview.metadata.name
            if self.preview.metadata and self.preview.metadata.name
            else ""
        )

    @property
    def metadata(self) -> PreviewMetadata | None:
        return self.preview.metadata if self.preview.metadata else None

    @property
    def spec(self) -> PreviewSpec | None:
        return self.preview.spec if self.preview.spec else None


class SyncSandboxPreviews:
    def __init__(self, sandbox: Sandbox):
        self.sandbox = sandbox

    @property
    def sandbox_name(self) -> str:
        return (
            self.sandbox.metadata.name
            if self.sandbox.metadata and self.sandbox.metadata.name
            else ""
        )

    def list(self) -> List[SyncSandboxPreview]:
        response = list_sandbox_previews(
            self.sandbox_name,
            client=client,
        )
        if response:
            return [SyncSandboxPreview(preview) for preview in response]
        raise errors.UnexpectedStatus(400, b"Failed to list previews")

    def create(self, preview: Union[Preview, Dict[str, Any]]) -> SyncSandboxPreview:
        if isinstance(preview, dict):
            preview = Preview.from_dict(preview)
        response = create_sandbox_preview(
            self.sandbox_name,
            body=preview,
            client=client,
        )
        if response:
            return SyncSandboxPreview(response)
        raise errors.UnexpectedStatus(400, b"Failed to create preview")

    def create_if_not_exists(self, preview: Union[Preview, Dict[str, Any]]) -> SyncSandboxPreview:
        if isinstance(preview, dict):
            preview = Preview.from_dict(preview)
        preview_name = preview.metadata.name if preview.metadata and preview.metadata.name else ""
        try:
            existing_preview = self.get(preview_name)
            return existing_preview
        except errors.UnexpectedStatus as e:
            if e.status_code == 404:
                return self.create(preview)
            raise e

    def get(self, preview_name: str) -> SyncSandboxPreview:
        response = get_sandbox_preview(
            self.sandbox_name,
            preview_name,
            client=client,
        )
        if response:
            return SyncSandboxPreview(response)
        raise errors.UnexpectedStatus(400, b"Failed to get preview")

    def delete(self, preview_name: str) -> Preview:
        response = delete_sandbox_preview(
            self.sandbox_name,
            preview_name,
            client=client,
        )
        if not response:
            raise errors.UnexpectedStatus(400, b"Failed to delete preview")

        # If the preview is in DELETING state, wait for it to be fully deleted
        if response.status == "DELETING":
            self._wait_for_deletion(preview_name)

        return response

    def _wait_for_deletion(self, preview_name: str, timeout_ms: int = 10000) -> None:
        """Wait for a preview to be fully deleted.

        Args:
            preview_name: Name of the preview to wait for
            timeout_ms: Timeout in milliseconds (default: 10000)

        Raises:
            Exception: If the preview is still in DELETING state after timeout
        """
        print(f"Waiting for preview deletion: {preview_name}")
        poll_interval = 0.5  # Poll every 500ms
        elapsed = 0.0
        timeout_seconds = timeout_ms / 1000.0

        while elapsed < timeout_seconds:
            try:
                response = get_sandbox_preview_detailed(
                    self.sandbox_name,
                    preview_name,
                    client=client,
                )
                if response.status_code == 404:
                    return
            except errors.UnexpectedStatus as e:
                # 404 means the preview is deleted
                if e.status_code == 404:
                    return
                raise
            # Preview still exists, wait and retry
            time.sleep(poll_interval)
            elapsed += poll_interval

        # Timeout reached, but deletion was initiated
        raise Exception(
            f"Preview deletion timeout: {preview_name} is still in DELETING state after {timeout_ms}ms"
        )


def to_utc_z(dt: datetime) -> str:
    iso_string = dt.isoformat()
    if iso_string.endswith("+00:00"):
        return iso_string.replace("+00:00", "Z")
    elif "T" in iso_string and not iso_string.endswith("Z"):
        return iso_string + "Z"
    return iso_string



from typing import Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="SandboxNetwork")


@_attrs_define
class SandboxNetwork:
    """Network configuration for a sandbox including egress IP binding. All three fields (vpcName, egressGatewayName,
    egressIpName) must be specified together to assign a dedicated IP.

        Attributes:
            egress_gateway_name (str): Name of the egress gateway in the VPC. Must be specified together with vpcName and
                egressIpName. Example: my-egress-gateway.
            vpc_name (str): Name of the VPC where the egress gateway is provisioned. Must be specified together with
                egressGatewayName and egressIpName. Example: my-vpc.
    """

    egress_gateway_name: str
    vpc_name: str
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        egress_gateway_name = self.egress_gateway_name

        vpc_name = self.vpc_name

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "egressGatewayName": egress_gateway_name,
                "vpcName": vpc_name,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T | None:
        if not src_dict:
            return None
        d = src_dict.copy()
        egress_gateway_name = (
            d.pop("egressGatewayName") if "egressGatewayName" in d else d.pop("egress_gateway_name")
        )

        vpc_name = d.pop("vpcName") if "vpcName" in d else d.pop("vpc_name")

        sandbox_network = cls(
            egress_gateway_name=egress_gateway_name,
            vpc_name=vpc_name,
        )

        sandbox_network.additional_properties = d
        return sandbox_network

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""LlamaIndex Model Integration Tests."""

pytest_plugins = []
import pytest  # noqa: E402

pytest.importorskip(
    "llama_index", reason="llama-index not installed (install with: blaxel[llamaindex])"
)

from llama_index.core.llms import ChatMessage  # type: ignore[import-not-found]

from blaxel.llamaindex import bl_model  # noqa: E402

TEST_MODELS = [
    "sandbox-openai",
]


@pytest.mark.asyncio(loop_scope="class")
class TestBlModel:
    """Test bl_model functionality."""

    @pytest.mark.parametrize("model_name", TEST_MODELS)
    async def test_can_chat_with_model(self, model_name: str):
        """Test chatting with a model."""
        model = await bl_model(model_name)
        result = await model.achat([ChatMessage(role="user", content="Say hello in one word")])

        assert result is not None
        assert result.message is not None
        assert result.message.content is not None



from typing import Any, Dict, List

from ..types import SandboxConfiguration
from .action import SandboxAction


class SandboxDrive(SandboxAction):
    """Asynchronous drive operations for mounting/unmounting drives to sandbox."""

    def __init__(self, sandbox_config: SandboxConfiguration):
        super().__init__(sandbox_config)

    async def mount(
        self,
        drive_name: str,
        mount_path: str,
        drive_path: str = "/",
    ) -> Dict[str, Any]:
        """
        Mount a drive to the sandbox at the specified mount path.

        Args:
            drive_name: Name of the drive to mount
            mount_path: Path in the sandbox where the drive should be mounted
            drive_path: Path within the drive to mount (default: "/")

        Returns:
            Dictionary with success status and mount information
        """
        if not mount_path.startswith("/"):
            mount_path = f"/{mount_path}"

        payload = {
            "driveName": drive_name,
            "mountPath": mount_path,
            "drivePath": drive_path,
        }

        client = self.get_client()
        response = await client.post("/drives/mount", json=payload)
        self.handle_response_error(response)
        return response.json()

    async def unmount(self, mount_path: str) -> Dict[str, Any]:
        """
        Unmount a drive from the sandbox.

        Args:
            mount_path: Path where the drive is currently mounted

        Returns:
            Dictionary with success status
        """
        if not mount_path.startswith("/"):
            mount_path = f"/{mount_path}"

        url_path = mount_path[1:]

        client = self.get_client()
        response = await client.delete(f"/drives/mount/{url_path}")
        self.handle_response_error(response)
        return response.json()

    async def list(self) -> List[Dict[str, Any]]:
        """
        List all currently mounted drives in the sandbox.

        Returns:
            List of dictionaries containing mount information for each drive
        """
        client = self.get_client()
        response = await client.get("/drives/mount")
        self.handle_response_error(response)

        result = response.json()
        if isinstance(result, dict) and "mounts" in result:
            return result["mounts"]
        return []



from http import HTTPStatus
from io import BytesIO
from typing import Any, Union

import httpx

from ... import errors
from ...client import Client
from ...models.directory import Directory
from ...models.error_response import ErrorResponse
from ...models.file_with_content import FileWithContent
from ...types import UNSET, File, Response, Unset


def _get_kwargs(
    path: str,
    *,
    download: Union[Unset, bool] = UNSET,
) -> dict[str, Any]:
    params: dict[str, Any] = {}

    params["download"] = download

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: dict[str, Any] = {
        "method": "get",
        "url": f"/filesystem/{path}",
        "params": params,
    }

    return _kwargs


def _parse_response(
    *, client: Client, response: httpx.Response
) -> Union[ErrorResponse, Union["Directory", "FileWithContent", File]] | None:
    if response.status_code == 200:

        def _parse_response_200(data: object) -> Union["Directory", "FileWithContent", File]:
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                response_200_type_0 = Directory.from_dict(data)

                return response_200_type_0
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                response_200_type_1 = FileWithContent.from_dict(data)

                return response_200_type_1
            except:  # noqa: E722
                pass
            if not isinstance(data, bytes):
                raise TypeError()
            response_200_type_2 = File(payload=BytesIO(data))

            return response_200_type_2

        response_200 = _parse_response_200(response.json())

        return response_200
    if response.status_code == 404:
        response_404 = ErrorResponse.from_dict(response.json())

        return response_404
    if response.status_code == 422:
        response_422 = ErrorResponse.from_dict(response.json())

        return response_422
    if response.status_code == 500:
        response_500 = ErrorResponse.from_dict(response.json())

        return response_500
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Client, response: httpx.Response
) -> Response[Union[ErrorResponse, Union["Directory", "FileWithContent", File]]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    path: str,
    *,
    client: Client,
    download: Union[Unset, bool] = UNSET,
) -> Response[Union[ErrorResponse, Union["Directory", "FileWithContent", File]]]:
    """Get file or directory information

     Get content of a file or listing of a directory. Use Accept header to control response format for
    files.

    Args:
        path (str):
        download (Union[Unset, bool]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[ErrorResponse, Union['Directory', 'FileWithContent', File]]]
    """

    kwargs = _get_kwargs(
        path=path,
        download=download,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    path: str,
    *,
    client: Client,
    download: Union[Unset, bool] = UNSET,
) -> Union[ErrorResponse, Union["Directory", "FileWithContent", File]] | None:
    """Get file or directory information

     Get content of a file or listing of a directory. Use Accept header to control response format for
    files.

    Args:
        path (str):
        download (Union[Unset, bool]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[ErrorResponse, Union['Directory', 'FileWithContent', File]]
    """

    return sync_detailed(
        path=path,
        client=client,
        download=download,
    ).parsed


async def asyncio_detailed(
    path: str,
    *,
    client: Client,
    download: Union[Unset, bool] = UNSET,
) -> Response[Union[ErrorResponse, Union["Directory", "FileWithContent", File]]]:
    """Get file or directory information

     Get content of a file or listing of a directory. Use Accept header to control response format for
    files.

    Args:
        path (str):
        download (Union[Unset, bool]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[ErrorResponse, Union['Directory', 'FileWithContent', File]]]
    """

    kwargs = _get_kwargs(
        path=path,
        download=download,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    path: str,
    *,
    client: Client,
    download: Union[Unset, bool] = UNSET,
) -> Union[ErrorResponse, Union["Directory", "FileWithContent", File]] | None:
    """Get file or directory information

     Get content of a file or listing of a directory. Use Accept header to control response format for
    files.

    Args:
        path (str):
        download (Union[Unset, bool]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[ErrorResponse, Union['Directory', 'FileWithContent', File]]
    """

    return (
        await asyncio_detailed(
            path=path,
            client=client,
            download=download,
        )
    ).parsed



from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.model_runtime_generation import ModelRuntimeGeneration
from ..models.model_runtime_type import ModelRuntimeType
from ..types import UNSET, Unset

T = TypeVar("T", bound="ModelRuntime")


@_attrs_define
class ModelRuntime:
    """Configuration identifying which external LLM provider and model this gateway endpoint proxies to

    Attributes:
        endpoint_name (Union[Unset, str]): Provider-specific endpoint name (e.g., HuggingFace Inference Endpoints name)
        generation (Union[Unset, ModelRuntimeGeneration]): Infrastructure generation. Empty (default) uses the classic
            deployment path. mk3 deploys through the model-gateway on microVM clusters.
        model (Union[Unset, str]): Model identifier at the provider (e.g., gpt-4.1, claude-sonnet-4-20250514, mistral-
            large-latest) Example: gpt-4.1.
        organization (Union[Unset, str]): Organization or account identifier at the provider (required for some
            providers like OpenAI) Example: org-abc123.
        type_ (Union[Unset, ModelRuntimeType]): LLM provider type determining the API protocol and authentication method
            Example: openai.
    """

    endpoint_name: Union[Unset, str] = UNSET
    generation: Union[Unset, ModelRuntimeGeneration] = UNSET
    model: Union[Unset, str] = UNSET
    organization: Union[Unset, str] = UNSET
    type_: Union[Unset, ModelRuntimeType] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        endpoint_name = self.endpoint_name

        generation: Union[Unset, str] = UNSET
        if not isinstance(self.generation, Unset):
            generation = self.generation.value

        model = self.model

        organization = self.organization

        type_: Union[Unset, str] = UNSET
        if not isinstance(self.type_, Unset):
            type_ = self.type_.value

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if endpoint_name is not UNSET:
            field_dict["endpointName"] = endpoint_name
        if generation is not UNSET:
            field_dict["generation"] = generation
        if model is not UNSET:
            field_dict["model"] = model
        if organization is not UNSET:
            field_dict["organization"] = organization
        if type_ is not UNSET:
            field_dict["type"] = type_

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T | None:
        if not src_dict:
            return None
        d = src_dict.copy()
        endpoint_name = d.pop("endpointName", d.pop("endpoint_name", UNSET))

        _generation = d.pop("generation", UNSET)
        generation: Union[Unset, ModelRuntimeGeneration]
        if isinstance(_generation, Unset):
            generation = UNSET
        else:
            generation = ModelRuntimeGeneration(_generation)

        model = d.pop("model", UNSET)

        organization = d.pop("organization", UNSET)

        _type_ = d.pop("type", d.pop("type_", UNSET))
        type_: Union[Unset, ModelRuntimeType]
        if isinstance(_type_, Unset):
            type_ = UNSET
        else:
            type_ = ModelRuntimeType(_type_)

        model_runtime = cls(
            endpoint_name=endpoint_name,
            generation=generation,
            model=model,
            organization=organization,
            type_=type_,
        )

        model_runtime.additional_properties = d
        return model_runtime

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.sandbox_error_details import SandboxErrorDetails


T = TypeVar("T", bound="SandboxError")


@_attrs_define
class SandboxError:
    """Error response returned by the CreateSandbox endpoint with extended details about the failure

    Attributes:
        code (str): Error code identifying the kind of error (e.g., INVALID_IMAGE, QUOTA_EXCEEDED) Example:
            INVALID_IMAGE.
        message (str): Human-readable error message describing what went wrong Example: Sandbox image blaxel/dev-ts-
            app:latest is not supported.
        details (Union[Unset, SandboxErrorDetails]): Additional error details. For INVALID_IMAGE errors, includes
            requested_image and supported_images array.
        sandbox_name (Union[Unset, str]): Name of the sandbox that failed to create Example: mysandbox.
        status_code (Union[Unset, int]): HTTP status code Example: 400.
        step (Union[Unset, str]): Processing step where the error occurred Example: validate_and_prepare.
        timestamp (Union[Unset, str]): ISO 8601 timestamp of when the error occurred Example:
            2025-12-19T22:36:29.336304095Z.
        workspace (Union[Unset, str]): Workspace name where the sandbox creation was attempted Example: main.
    """

    code: str
    message: str
    details: Union[Unset, "SandboxErrorDetails"] = UNSET
    sandbox_name: Union[Unset, str] = UNSET
    status_code: Union[Unset, int] = UNSET
    step: Union[Unset, str] = UNSET
    timestamp: Union[Unset, str] = UNSET
    workspace: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        code = self.code

        message = self.message

        details: Union[Unset, dict[str, Any]] = UNSET
        if (
            self.details
            and not isinstance(self.details, Unset)
            and not isinstance(self.details, dict)
        ):
            details = self.details.to_dict()
        elif self.details and isinstance(self.details, dict):
            details = self.details

        sandbox_name = self.sandbox_name

        status_code = self.status_code

        step = self.step

        timestamp = self.timestamp

        workspace = self.workspace

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "code": code,
                "message": message,
            }
        )
        if details is not UNSET:
            field_dict["details"] = details
        if sandbox_name is not UNSET:
            field_dict["sandbox_name"] = sandbox_name
        if status_code is not UNSET:
            field_dict["status_code"] = status_code
        if step is not UNSET:
            field_dict["step"] = step
        if timestamp is not UNSET:
            field_dict["timestamp"] = timestamp
        if workspace is not UNSET:
            field_dict["workspace"] = workspace

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T | None:
        from ..models.sandbox_error_details import SandboxErrorDetails

        if not src_dict:
            return None
        d = src_dict.copy()
        code = d.pop("code")

        message = d.pop("message")

        _details = d.pop("details", UNSET)
        details: Union[Unset, SandboxErrorDetails]
        if isinstance(_details, Unset):
            details = UNSET
        else:
            details = SandboxErrorDetails.from_dict(_details)

        sandbox_name = d.pop("sandbox_name", UNSET)

        status_code = d.pop("status_code", UNSET)

        step = d.pop("step", UNSET)

        timestamp = d.pop("timestamp", UNSET)

        workspace = d.pop("workspace", UNSET)

        sandbox_error = cls(
            code=code,
            message=message,
            details=details,
            sandbox_name=sandbox_name,
            status_code=status_code,
            step=step,
            timestamp=timestamp,
            workspace=workspace,
        )

        sandbox_error.additional_properties = d
        return sandbox_error

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.function_runtime_generation import FunctionRuntimeGeneration
from ..models.function_runtime_transport import FunctionRuntimeTransport
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.env import Env


T = TypeVar("T", bound="FunctionRuntime")


@_attrs_define
class FunctionRuntime:
    """Runtime configuration defining how the MCP server function is deployed and scaled

    Attributes:
        envs (Union[Unset, list['Env']]): Environment variables injected into the function. Supports Kubernetes EnvVar
            format with valueFrom references.
        generation (Union[Unset, FunctionRuntimeGeneration]): Infrastructure generation: mk2 (containers, 2-10s cold
            starts, 15+ global regions) or mk3 (microVMs, sub-25ms cold starts) Example: mk3.
        image (Union[Unset, str]): Container image built by Blaxel when deploying with 'bl deploy'. This field is auto-
            populated during deployment.
        max_scale (Union[Unset, int]): Maximum number of concurrent function instances for auto-scaling Example: 10.
        memory (Union[Unset, int]): Memory allocation in megabytes. Also determines CPU allocation (CPU cores = memory
            in MB / 2048, e.g., 4096MB = 2 CPUs). Example: 2048.
        min_scale (Union[Unset, int]): Minimum instances to keep warm. Set to 1+ to eliminate cold starts, 0 for scale-
            to-zero.
        transport (Union[Unset, FunctionRuntimeTransport]): Transport compatibility for the MCP, can be "websocket" or
            "http-stream" Example: http-stream.
    """

    envs: Union[Unset, list["Env"]] = UNSET
    generation: Union[Unset, FunctionRuntimeGeneration] = UNSET
    image: Union[Unset, str] = UNSET
    max_scale: Union[Unset, int] = UNSET
    memory: Union[Unset, int] = UNSET
    min_scale: Union[Unset, int] = UNSET
    transport: Union[Unset, FunctionRuntimeTransport] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        envs: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.envs, Unset):
            envs = []
            for envs_item_data in self.envs:
                if type(envs_item_data) is dict:
                    envs_item = envs_item_data
                else:
                    envs_item = envs_item_data.to_dict()
                envs.append(envs_item)

        generation: Union[Unset, str] = UNSET
        if not isinstance(self.generation, Unset):
            generation = self.generation.value

        image = self.image

        max_scale = self.max_scale

        memory = self.memory

        min_scale = self.min_scale

        transport: Union[Unset, str] = UNSET
        if not isinstance(self.transport, Unset):
            transport = self.transport.value

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if envs is not UNSET:
            field_dict["envs"] = envs
        if generation is not UNSET:
            field_dict["generation"] = generation
        if image is not UNSET:
            field_dict["image"] = image
        if max_scale is not UNSET:
            field_dict["maxScale"] = max_scale
        if memory is not UNSET:
            field_dict["memory"] = memory
        if min_scale is not UNSET:
            field_dict["minScale"] = min_scale
        if transport is not UNSET:
            field_dict["transport"] = transport

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T | None:
        from ..models.env import Env

        if not src_dict:
            return None
        d = src_dict.copy()
        envs = []
        _envs = d.pop("envs", UNSET)
        for envs_item_data in _envs or []:
            envs_item = Env.from_dict(envs_item_data)

            envs.append(envs_item)

        _generation = d.pop("generation", UNSET)
        generation: Union[Unset, FunctionRuntimeGeneration]
        if isinstance(_generation, Unset):
            generation = UNSET
        else:
            generation = FunctionRuntimeGeneration(_generation)

        image = d.pop("image", UNSET)

        max_scale = d.pop("maxScale", d.pop("max_scale", UNSET))

        memory = d.pop("memory", UNSET)

        min_scale = d.pop("minScale", d.pop("min_scale", UNSET))

        _transport = d.pop("transport", UNSET)
        transport: Union[Unset, FunctionRuntimeTransport]
        if isinstance(_transport, Unset):
            transport = UNSET
        else:
            transport = FunctionRuntimeTransport(_transport)

        function_runtime = cls(
            envs=envs,
            generation=generation,
            image=image,
            max_scale=max_scale,
            memory=memory,
            min_scale=min_scale,
            transport=transport,
        )

        function_runtime.additional_properties = d
        return function_runtime

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.test_feature_flag_response_200_payload import TestFeatureFlagResponse200Payload


T = TypeVar("T", bound="TestFeatureFlagResponse200")


@_attrs_define
class TestFeatureFlagResponse200:
    """
    Attributes:
        enabled (Union[Unset, bool]):
        evaluated_at (Union[Unset, datetime.datetime]):
        feature_key (Union[Unset, str]):
        payload (Union[Unset, TestFeatureFlagResponse200Payload]):
        variant (Union[Unset, str]):
    """

    enabled: Union[Unset, bool] = UNSET
    evaluated_at: Union[Unset, datetime.datetime] = UNSET
    feature_key: Union[Unset, str] = UNSET
    payload: Union[Unset, "TestFeatureFlagResponse200Payload"] = UNSET
    variant: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        enabled = self.enabled

        evaluated_at: Union[Unset, str] = UNSET
        if not isinstance(self.evaluated_at, Unset):
            evaluated_at = self.evaluated_at.isoformat()

        feature_key = self.feature_key

        payload: Union[Unset, dict[str, Any]] = UNSET
        if (
            self.payload
            and not isinstance(self.payload, Unset)
            and not isinstance(self.payload, dict)
        ):
            payload = self.payload.to_dict()
        elif self.payload and isinstance(self.payload, dict):
            payload = self.payload

        variant = self.variant

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if enabled is not UNSET:
            field_dict["enabled"] = enabled
        if evaluated_at is not UNSET:
            field_dict["evaluatedAt"] = evaluated_at
        if feature_key is not UNSET:
            field_dict["featureKey"] = feature_key
        if payload is not UNSET:
            field_dict["payload"] = payload
        if variant is not UNSET:
            field_dict["variant"] = variant

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T | None:
        from ..models.test_feature_flag_response_200_payload import (
            TestFeatureFlagResponse200Payload,
        )

        if not src_dict:
            return None
        d = src_dict.copy()
        enabled = d.pop("enabled", UNSET)

        _evaluated_at = d.pop("evaluatedAt", d.pop("evaluated_at", UNSET))
        evaluated_at: Union[Unset, datetime.datetime]
        if isinstance(_evaluated_at, Unset):
            evaluated_at = UNSET
        else:
            evaluated_at = isoparse(_evaluated_at)

        feature_key = d.pop("featureKey", d.pop("feature_key", UNSET))

        _payload = d.pop("payload", UNSET)
        payload: Union[Unset, TestFeatureFlagResponse200Payload]
        if isinstance(_payload, Unset):
            payload = UNSET
        else:
            payload = TestFeatureFlagResponse200Payload.from_dict(_payload)

        variant = d.pop("variant", UNSET)

        test_feature_flag_response_200 = cls(
            enabled=enabled,
            evaluated_at=evaluated_at,
            feature_key=feature_key,
            payload=payload,
            variant=variant,
        )

        test_feature_flag_response_200.additional_properties = d
        return test_feature_flag_response_200

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class JobExecutionStatus(str, Enum):
    CANCELLED = "cancelled"
    CANCELLING = "cancelling"
    FAILED = "failed"
    PENDING = "pending"
    QUEUED = "queued"
    RUNNING = "running"
    SUCCEEDED = "succeeded"
    TIMEOUT = "timeout"

    def __str__(self) -> str:
        return str(self.value)

    @classmethod
    def _missing_(cls, value: object) -> "JobExecutionStatus | None":
        if isinstance(value, str):
            upper_value = value.upper()
            for member in cls:
                if member.value.upper() == upper_value:
                    return member
        return None



"""Pytest configuration for integration tests."""

import pytest
import pytest_asyncio


@pytest_asyncio.fixture(autouse=True, scope="class", loop_scope="class")
async def reset_client():
    """Reset the global client's async httpx client for each test class.

    This ensures each test class gets a fresh httpx.AsyncClient bound to
    its own event loop, preventing "Event loop is closed" errors.
    """
    from blaxel.core.client.client import client

    # Reset at the start of each test class
    client._async_client = None

    yield

    # Reset at the end (the event loop will close after this)
    client._async_client = None


# Mark all tests in this module as integration tests
def pytest_collection_modifyitems(config, items):
    """Add integration marker to all tests in this directory."""
    for item in items:
        if "integration" in str(item.fspath):
            item.add_marker(pytest.mark.integration)



from http import HTTPStatus
from typing import Any, Union

import httpx

from ... import errors
from ...client import Client
from ...models.error import Error
from ...models.function import Function
from ...types import UNSET, Response, Unset


def _get_kwargs(
    function_name: str,
    *,
    show_secrets: Union[Unset, bool] = UNSET,
) -> dict[str, Any]:
    params: dict[str, Any] = {}

    params["show_secrets"] = show_secrets

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: dict[str, Any] = {
        "method": "get",
        "url": f"/functions/{function_name}",
        "params": params,
    }

    return _kwargs


def _parse_response(*, client: Client, response: httpx.Response) -> Union[Error, Function] | None:
    if response.status_code == 200:
        response_200 = Function.from_dict(response.json())

        return response_200
    if response.status_code == 401:
        response_401 = Error.from_dict(response.json())

        return response_401
    if response.status_code == 403:
        response_403 = Error.from_dict(response.json())

        return response_403
    if response.status_code == 404:
        response_404 = Error.from_dict(response.json())

        return response_404
    if response.status_code == 500:
        response_500 = Error.from_dict(response.json())

        return response_500
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Client, response: httpx.Response
) -> Response[Union[Error, Function]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    function_name: str,
    *,
    client: Client,
    show_secrets: Union[Unset, bool] = UNSET,
) -> Response[Union[Error, Function]]:
    """Get MCP server

     Returns detailed information about an MCP server function including its deployment status, available
    tools, transport configuration, and endpoint URL.

    Args:
        function_name (str):
        show_secrets (Union[Unset, bool]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Error, Function]]
    """

    kwargs = _get_kwargs(
        function_name=function_name,
        show_secrets=show_secrets,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    function_name: str,
    *,
    client: Client,
    show_secrets: Union[Unset, bool] = UNSET,
) -> Union[Error, Function] | None:
    """Get MCP server

     Returns detailed information about an MCP server function including its deployment status, available
    tools, transport configuration, and endpoint URL.

    Args:
        function_name (str):
        show_secrets (Union[Unset, bool]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Error, Function]
    """

    return sync_detailed(
        function_name=function_name,
        client=client,
        show_secrets=show_secrets,
    ).parsed


async def asyncio_detailed(
    function_name: str,
    *,
    client: Client,
    show_secrets: Union[Unset, bool] = UNSET,
) -> Response[Union[Error, Function]]:
    """Get MCP server

     Returns detailed information about an MCP server function including its deployment status, available
    tools, transport configuration, and endpoint URL.

    Args:
        function_name (str):
        show_secrets (Union[Unset, bool]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Error, Function]]
    """

    kwargs = _get_kwargs(
        function_name=function_name,
        show_secrets=show_secrets,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    function_name: str,
    *,
    client: Client,
    show_secrets: Union[Unset, bool] = UNSET,
) -> Union[Error, Function] | None:
    """Get MCP server

     Returns detailed information about an MCP server function including its deployment status, available
    tools, transport configuration, and endpoint URL.

    Args:
        function_name (str):
        show_secrets (Union[Unset, bool]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Error, Function]
    """

    return (
        await asyncio_detailed(
            function_name=function_name,
            client=client,
            show_secrets=show_secrets,
        )
    ).parsed



"""Google ADK Model Integration Tests."""

pytest_plugins = []
import pytest  # noqa: E402

pytest.importorskip(
    "google.adk", reason="google-adk not installed (install with: blaxel[googleadk])"
)

from google.adk.models import LlmRequest  # noqa: E402
from google.genai import types  # noqa: E402

from blaxel.googleadk import bl_model  # noqa: E402

TEST_MODELS = [
    "sandbox-openai",
]


@pytest.mark.asyncio(loop_scope="class")
class TestBlModel:
    """Test bl_model functionality."""

    @pytest.mark.parametrize("model_name", TEST_MODELS)
    async def test_can_create_model(self, model_name: str):
        """Test creating a model."""
        model = await bl_model(model_name)

        assert model is not None

    @pytest.mark.parametrize("model_name", TEST_MODELS)
    async def test_can_call_model(self, model_name: str):
        """Test making an actual request to the model."""
        model = await bl_model(model_name)

        request = LlmRequest(
            contents=[
                types.Content(
                    parts=[types.Part(text="Say hello in one word")],
                    role="user",
                )
            ],
            config=types.GenerateContentConfig(),
        )

        collected_text = ""
        async for response in model.generate_content_async(request):
            assert response is not None
            if response.content and response.content.parts:
                for part in response.content.parts:
                    if part.text:
                        collected_text += part.text

        assert len(collected_text) > 0



from http import HTTPStatus
from typing import Any, Union

import httpx

from ... import errors
from ...client import Client
from ...models.error import Error
from ...models.sandbox import Sandbox
from ...types import Response


def _get_kwargs(
    sandbox_name: str,
    *,
    body: Sandbox,
) -> dict[str, Any]:
    headers: dict[str, Any] = {}

    _kwargs: dict[str, Any] = {
        "method": "put",
        "url": f"/sandboxes/{sandbox_name}",
    }

    if type(body) is dict:
        _body = body
    else:
        _body = body.to_dict()

    _kwargs["json"] = _body
    headers["Content-Type"] = "application/json"

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(*, client: Client, response: httpx.Response) -> Union[Error, Sandbox] | None:
    if response.status_code == 200:
        response_200 = Sandbox.from_dict(response.json())

        return response_200
    if response.status_code == 400:
        response_400 = Error.from_dict(response.json())

        return response_400
    if response.status_code == 401:
        response_401 = Error.from_dict(response.json())

        return response_401
    if response.status_code == 403:
        response_403 = Error.from_dict(response.json())

        return response_403
    if response.status_code == 404:
        response_404 = Error.from_dict(response.json())

        return response_404
    if response.status_code == 500:
        response_500 = Error.from_dict(response.json())

        return response_500
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(*, client: Client, response: httpx.Response) -> Response[Union[Error, Sandbox]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    sandbox_name: str,
    *,
    client: Client,
    body: Sandbox,
) -> Response[Union[Error, Sandbox]]:
    """Update sandbox

     Updates a sandbox's configuration. Note that certain changes (like image or memory) may reset the
    sandbox state. Use lifecycle policies to control automatic cleanup.

    Args:
        sandbox_name (str):
        body (Sandbox): Lightweight virtual machine for secure AI code execution. Sandboxes resume
            from standby in under 25ms and automatically scale to zero after inactivity, preserving
            memory state including running processes and filesystem.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Error, Sandbox]]
    """

    kwargs = _get_kwargs(
        sandbox_name=sandbox_name,
        body=body,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    sandbox_name: str,
    *,
    client: Client,
    body: Sandbox,
) -> Union[Error, Sandbox] | None:
    """Update sandbox

     Updates a sandbox's configuration. Note that certain changes (like image or memory) may reset the
    sandbox state. Use lifecycle policies to control automatic cleanup.

    Args:
        sandbox_name (str):
        body (Sandbox): Lightweight virtual machine for secure AI code execution. Sandboxes resume
            from standby in under 25ms and automatically scale to zero after inactivity, preserving
            memory state including running processes and filesystem.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Error, Sandbox]
    """

    return sync_detailed(
        sandbox_name=sandbox_name,
        client=client,
        body=body,
    ).parsed


async def asyncio_detailed(
    sandbox_name: str,
    *,
    client: Client,
    body: Sandbox,
) -> Response[Union[Error, Sandbox]]:
    """Update sandbox

     Updates a sandbox's configuration. Note that certain changes (like image or memory) may reset the
    sandbox state. Use lifecycle policies to control automatic cleanup.

    Args:
        sandbox_name (str):
        body (Sandbox): Lightweight virtual machine for secure AI code execution. Sandboxes resume
            from standby in under 25ms and automatically scale to zero after inactivity, preserving
            memory state including running processes and filesystem.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Error, Sandbox]]
    """

    kwargs = _get_kwargs(
        sandbox_name=sandbox_name,
        body=body,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    sandbox_name: str,
    *,
    client: Client,
    body: Sandbox,
) -> Union[Error, Sandbox] | None:
    """Update sandbox

     Updates a sandbox's configuration. Note that certain changes (like image or memory) may reset the
    sandbox state. Use lifecycle policies to control automatic cleanup.

    Args:
        sandbox_name (str):
        body (Sandbox): Lightweight virtual machine for secure AI code execution. Sandboxes resume
            from standby in under 25ms and automatically scale to zero after inactivity, preserving
            memory state including running processes and filesystem.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Error, Sandbox]
    """

    return (
        await asyncio_detailed(
            sandbox_name=sandbox_name,
            client=client,
            body=body,
        )
    ).parsed



from __future__ import annotations

import os
from logging import getLogger
from typing import TYPE_CHECKING, Any, Dict, List, Sequence, Union

from blaxel.core import bl_model as bl_model_core
from blaxel.core import settings

# Transformers is a dependency of DeepSeek, and it logs a lot of warnings that are not useful
os.environ["TRANSFORMERS_NO_ADVISORY_WARNINGS"] = "1"

if TYPE_CHECKING:
    from llama_index.core.base.llms.types import (  # type: ignore[import-not-found]
        ChatMessage,
        ChatResponse,
        ChatResponseAsyncGen,
        ChatResponseGen,
        CompletionResponse,
        CompletionResponseAsyncGen,
        CompletionResponseGen,
    )
    from llama_index.core.llms.llm import (  # type: ignore[import-not-found]
        ToolSelection,
    )
    from llama_index.core.tools.types import BaseTool  # type: ignore[import-not-found]

# Runtime imports needed for class inheritance and construction
from llama_index.core.base.llms.types import (  # type: ignore[import-not-found]
    LLMMetadata,
)
from llama_index.core.llms.function_calling import (  # type: ignore[import-not-found]
    FunctionCallingLLM,
)
from pydantic import PrivateAttr  # type: ignore[import-not-found]

logger = getLogger(__name__)

DEFAULT_CONTEXT_WINDOW = 128000
DEFAULT_NUM_OUTPUT = 4096


class TokenRefreshingLLM(FunctionCallingLLM):
    """Wrapper for LlamaIndex LLMs that refreshes token before each call.

    Inherits from FunctionCallingLLM to maintain type compatibility with
    LlamaIndex's agents and components that validate isinstance(model, LLM).
    """

    _model_config_data: dict = PrivateAttr(default_factory=dict)
    _wrapped: Any = PrivateAttr(default=None)

    def __init__(self, model_config: dict):
        super().__init__()
        self._model_config_data = model_config
        self._wrapped = self._create_model()

    @classmethod
    def class_name(cls) -> str:
        return "TokenRefreshingLLM"

    @property
    def wrapped_model(self) -> Any:
        """Access the underlying wrapped LLM model."""
        return self._wrapped

    @property
    def metadata(self) -> LLMMetadata:
        """Get LLM metadata, with fallback for unknown model names."""
        try:
            return self._wrapped.metadata
        except (ValueError, KeyError) as e:
            logger.warning(f"Could not get metadata from wrapped model: {e}. Using defaults.")
            return LLMMetadata(
                context_window=DEFAULT_CONTEXT_WINDOW,
                num_output=DEFAULT_NUM_OUTPUT,
                is_chat_model=True,
                model_name=self._model_config_data.get("model", "unknown"),
            )

    def _create_model(self):
        """Create the model instance with current token."""
        config = self._model_config_data
        model_type = config["type"]
        model = config["model"]
        url = config["url"]
        kwargs = config.get("kwargs", {})

        if model_type == "anthropic":
            from llama_index.llms.anthropic import (  # type: ignore[import-not-found]
                Anthropic,
            )

            return Anthropic(
                model=model,
                api_key=settings.auth.token,
                base_url=url,
                default_headers=settings.auth.get_headers(),
                **kwargs,
            )
        elif model_type == "xai":
            from llama_index.llms.groq import Groq  # type: ignore[import-not-found]

            return Groq(
                model=model,
                api_key=settings.auth.token,
                api_base=f"{url}/v1",
                **kwargs,
            )
        elif model_type == "gemini":
            from google.genai.types import HttpOptions
            from llama_index.llms.google_genai import (  # type: ignore[import-not-found]
                GoogleGenAI,
            )

            return GoogleGenAI(
                api_key=settings.auth.token,
                model=model,
                api_base=f"{url}/v1",
                http_options=HttpOptions(
                    base_url=url,
                    headers=settings.auth.get_headers(),
                ),
                **kwargs,
            )
        elif model_type == "cohere":
            from .custom.cohere import Cohere  # type: ignore[import-not-found]

            return Cohere(model=model, api_key=settings.auth.token, api_base=url, **kwargs)
        elif model_type == "deepseek":
            from llama_index.llms.deepseek import (  # type: ignore[import-not-found]
                DeepSeek,
            )

            return DeepSeek(
                model=model,
                api_key=settings.auth.token,
                api_base=f"{url}/v1",
                **kwargs,
            )
        elif model_type == "mistral":
            from llama_index.llms.mistralai import (  # type: ignore[import-not-found]
                MistralAI,
            )

            return MistralAI(model=model, api_key=settings.auth.token, endpoint=url, **kwargs)
        elif model_type == "cerebras":
            from llama_index.llms.cerebras import (  # type: ignore[import-not-found]
                Cerebras,
            )

            return Cerebras(
                model=model,
                api_key=settings.auth.token,
                api_base=f"{url}/v1",
                **kwargs,
            )
        else:
            from llama_index.llms.openai import OpenAI  # type: ignore[import-not-found]

            if model_type != "openai":
                logger.warning(
                    f"Model {model} is not supported by LlamaIndex, defaulting to OpenAI"
                )

            return OpenAI(
                model=model,
                api_key=settings.auth.token,
                api_base=f"{url}/v1",
                **kwargs,
            )

    def _refresh_token(self):
        """Refresh the token and recreate the model if needed."""
        current_token = settings.auth.token

        new_token = settings.auth.token

        if current_token != new_token:
            self._wrapped = self._create_model()

    # --- Core LLM methods with token refresh ---

    def chat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponse:
        self._refresh_token()
        return self._wrapped.chat(messages, **kwargs)

    async def achat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponse:
        self._refresh_token()
        return await self._wrapped.achat(messages, **kwargs)

    def complete(self, prompt: str, formatted: bool = False, **kwargs: Any) -> CompletionResponse:
        self._refresh_token()
        return self._wrapped.complete(prompt, formatted=formatted, **kwargs)

    async def acomplete(
        self, prompt: str, formatted: bool = False, **kwargs: Any
    ) -> CompletionResponse:
        self._refresh_token()
        return await self._wrapped.acomplete(prompt, formatted=formatted, **kwargs)

    def stream_chat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponseGen:
        self._refresh_token()
        return self._wrapped.stream_chat(messages, **kwargs)

    async def astream_chat(
        self, messages: Sequence[ChatMessage], **kwargs: Any
    ) -> ChatResponseAsyncGen:
        self._refresh_token()
        result = self._wrapped.astream_chat(messages, **kwargs)
        # Handle both coroutine and async generator patterns
        if hasattr(result, "__aiter__"):
            return result
        return await result

    def stream_complete(
        self, prompt: str, formatted: bool = False, **kwargs: Any
    ) -> CompletionResponseGen:
        self._refresh_token()
        return self._wrapped.stream_complete(prompt, formatted=formatted, **kwargs)

    async def astream_complete(
        self, prompt: str, formatted: bool = False, **kwargs: Any
    ) -> CompletionResponseAsyncGen:
        self._refresh_token()
        result = self._wrapped.astream_complete(prompt, formatted=formatted, **kwargs)
        # Handle both coroutine and async generator patterns
        if hasattr(result, "__aiter__"):
            return result
        return await result

    # --- FunctionCallingLLM methods (delegate to wrapped model) ---

    def _prepare_chat_with_tools(
        self,
        tools: Sequence[BaseTool],
        user_msg: Union[str, ChatMessage, None] = None,
        chat_history: List[ChatMessage] | None = None,
        verbose: bool = False,
        allow_parallel_tool_calls: bool = False,
        tool_required: Any = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        if hasattr(self._wrapped, "_prepare_chat_with_tools"):
            return self._wrapped._prepare_chat_with_tools(
                tools,
                user_msg=user_msg,
                chat_history=chat_history,
                verbose=verbose,
                allow_parallel_tool_calls=allow_parallel_tool_calls,
                tool_required=tool_required,
                **kwargs,
            )
        raise NotImplementedError(
            f"The wrapped model ({type(self._wrapped).__name__}) does not support function calling"
        )

    def get_tool_calls_from_response(
        self,
        response: ChatResponse,
        error_on_no_tool_call: bool = True,
        **kwargs: Any,
    ) -> List[ToolSelection]:
        if hasattr(self._wrapped, "get_tool_calls_from_response"):
            return self._wrapped.get_tool_calls_from_response(
                response,
                error_on_no_tool_call=error_on_no_tool_call,
                **kwargs,
            )
        raise NotImplementedError(
            f"The wrapped model ({type(self._wrapped).__name__}) does not support function calling"
        )


async def bl_model(name, **kwargs):
    url, type, model = await bl_model_core(name).get_parameters()

    # Store model configuration for recreation
    model_config = {"type": type, "model": model, "url": url, "kwargs": kwargs}

    # Create and return the wrapper
    return TokenRefreshingLLM(model_config)



import asyncio

import pytest
import pytest_asyncio

from blaxel.core.sandbox import SandboxInstance
from tests.helpers import async_sleep, default_image, default_labels, unique_name


@pytest.mark.asyncio(loop_scope="class")
class TestFilesystemOperations:
    """Test sandbox filesystem operations."""

    sandbox: SandboxInstance
    sandbox_name: str

    @pytest_asyncio.fixture(autouse=True, scope="class", loop_scope="class")
    async def setup_sandbox(self, request):
        """Set up a sandbox for the test class."""
        request.cls.sandbox_name = unique_name("fs-test")
        request.cls.sandbox = await SandboxInstance.create(
            {
                "name": request.cls.sandbox_name,
                "image": default_image,
                "memory": 2048,
                "labels": default_labels,
            }
        )

        yield

        # Cleanup
        try:
            await self.sandbox.delete()
        except Exception:
            pass


@pytest.mark.asyncio(loop_scope="class")
class TestWriteAndRead(TestFilesystemOperations):
    """Test write and read operations."""

    async def test_writes_and_reads_text_file(self):
        """Test writing and reading a text file."""
        content = "Hello, World!"
        path = "/tmp/test-write.txt"

        await self.sandbox.fs.write(path, content)
        result = await self.sandbox.fs.read(path)

        assert result == content

    async def test_writes_and_reads_unicode_content(self):
        """Test writing and reading unicode content."""
        content = "Hello 世界 🌍 émojis"
        path = "/tmp/test-unicode.txt"

        await self.sandbox.fs.write(path, content)
        result = await self.sandbox.fs.read(path)

        assert result == content

    async def test_writes_and_reads_multiline_content(self):
        """Test writing and reading multiline content."""
        content = "Line 1\nLine 2\nLine 3"
        path = "/tmp/test-multiline.txt"

        await self.sandbox.fs.write(path, content)
        result = await self.sandbox.fs.read(path)

        assert result == content

    async def test_overwrites_existing_file(self):
        """Test overwriting an existing file."""
        path = "/tmp/test-overwrite.txt"

        await self.sandbox.fs.write(path, "original")
        await self.sandbox.fs.write(path, "updated")

        result = await self.sandbox.fs.read(path)
        assert result == "updated"


@pytest.mark.asyncio(loop_scope="class")
class TestBinaryOperations(TestFilesystemOperations):
    """Test binary write and read operations."""

    async def test_read_binary_works_on_text_files(self):
        """Test that readBinary works on text files too."""
        path = "/tmp/test-binary-text.txt"
        await self.sandbox.fs.write(path, "text content")

        blob = await self.sandbox.fs.read_binary(path)
        assert isinstance(blob, bytes)

        text = blob.decode("utf-8")
        assert text == "text content"


@pytest.mark.asyncio(loop_scope="class")
class TestWriteBinary(TestFilesystemOperations):
    """Test write_binary operations."""

    async def test_writes_and_reads_binary_content(self):
        """Test writing and reading binary content."""
        import time

        path = f"/tmp/binary-test-{int(time.time() * 1000)}.bin"
        # Create binary content with various byte values
        binary_content = bytes([0, 1, 2, 127, 128, 255, 0, 100, 200])

        await self.sandbox.fs.write_binary(path, binary_content)
        result = await self.sandbox.fs.read_binary(path)

        assert result == binary_content

    async def test_writes_binary_from_bytearray(self):
        """Test writing binary content from bytearray."""
        import time

        path = f"/tmp/bytearray-test-{int(time.time() * 1000)}.bin"
        binary_content = bytearray([10, 20, 30, 40, 50])

        await self.sandbox.fs.write_binary(path, binary_content)
        result = await self.sandbox.fs.read_binary(path)

        assert result == bytes(binary_content)

    async def test_writes_large_binary_file(self):
        """Test writing a large binary file."""
        import time

        path = f"/tmp/large-binary-{int(time.time() * 1000)}.bin"
        # Create ~500KB of binary data
        binary_content = bytes(range(256)) * 2000

        await self.sandbox.fs.write_binary(path, binary_content)
        result = await self.sandbox.fs.read_binary(path)

        assert len(result) == len(binary_content)
        assert result == binary_content


@pytest.mark.asyncio(loop_scope="class")
class TestWriteTree(TestFilesystemOperations):
    """Test write_tree operations."""

    async def test_writes_multiple_files(self):
        """Test writing multiple files in a tree structure."""
        import time

        base_dir = f"/tmp/tree-test-{int(time.time() * 1000)}"
        files = [
            {"path": f"{base_dir}/file1.txt", "content": "content1"},
            {"path": f"{base_dir}/file2.txt", "content": "content2"},
            {"path": f"{base_dir}/subdir/file3.txt", "content": "content3"},
        ]

        result = await self.sandbox.fs.write_tree(files)
        assert result is not None

        # Verify files were created
        content1 = await self.sandbox.fs.read(f"{base_dir}/file1.txt")
        content2 = await self.sandbox.fs.read(f"{base_dir}/file2.txt")
        content3 = await self.sandbox.fs.read(f"{base_dir}/subdir/file3.txt")

        assert content1 == "content1"
        assert content2 == "content2"
        assert content3 == "content3"

    async def test_writes_tree_with_destination_path(self):
        """Test writing files with a destination path prefix."""
        import time

        base_dir = f"/tmp/tree-dest-{int(time.time() * 1000)}"
        files = [
            {"path": "a.txt", "content": "file a"},
            {"path": "b.txt", "content": "file b"},
        ]

        result = await self.sandbox.fs.write_tree(files, destination_path=base_dir)
        assert result is not None

        # Verify files were created at destination
        content_a = await self.sandbox.fs.read(f"{base_dir}/a.txt")
        content_b = await self.sandbox.fs.read(f"{base_dir}/b.txt")

        assert content_a == "file a"
        assert content_b == "file b"


@pytest.mark.asyncio(loop_scope="class")
class TestDownload(TestFilesystemOperations):
    """Test download operations."""

    async def test_downloads_file_to_local(self):
        """Test downloading a file from sandbox to local filesystem."""
        import os
        import tempfile
        import time

        # Create a file in the sandbox
        sandbox_path = f"/tmp/download-test-{int(time.time() * 1000)}.txt"
        content = "Download me!"
        await self.sandbox.fs.write(sandbox_path, content)

        # Download to local temp file
        with tempfile.NamedTemporaryFile(delete=False, suffix=".txt") as tmp:
            local_path = tmp.name

        try:
            await self.sandbox.fs.download(sandbox_path, local_path)

            # Verify local file content
            with open(local_path) as f:
                local_content = f.read()

            assert local_content == content
        finally:
            # Cleanup
            if os.path.exists(local_path):
                os.remove(local_path)

    async def test_downloads_binary_file_to_local(self):
        """Test downloading a binary file from sandbox to local filesystem."""
        import os
        import tempfile
        import time

        # Create a binary file in the sandbox
        sandbox_path = f"/tmp/download-binary-{int(time.time() * 1000)}.bin"
        binary_content = bytes([0, 1, 2, 127, 128, 255])
        await self.sandbox.fs.write_binary(sandbox_path, binary_content)

        # Download to local temp file
        with tempfile.NamedTemporaryFile(delete=False, suffix=".bin") as tmp:
            local_path = tmp.name

        try:
            await self.sandbox.fs.download(sandbox_path, local_path)

            # Verify local file content
            with open(local_path, "rb") as f:
                local_content = f.read()

            assert local_content == binary_content
        finally:
            # Cleanup
            if os.path.exists(local_path):
                os.remove(local_path)


@pytest.mark.asyncio(loop_scope="class")
class TestListDirectory(TestFilesystemOperations):
    """Test ls (list directory) operations."""

    async def test_lists_files_in_directory(self):
        """Test listing files in a directory."""
        # Create some test files
        await self.sandbox.fs.write("/tmp/ls-test/file1.txt", "content1")
        await self.sandbox.fs.write("/tmp/ls-test/file2.txt", "content2")

        listing = await self.sandbox.fs.ls("/tmp/ls-test")

        assert listing.files is not None
        assert len(listing.files) >= 2

        names = [f.name for f in listing.files]
        assert "file1.txt" in names
        assert "file2.txt" in names

    async def test_lists_subdirectories(self):
        """Test listing subdirectories."""
        await self.sandbox.fs.mkdir("/tmp/ls-subdir-test/subdir1")
        await self.sandbox.fs.mkdir("/tmp/ls-subdir-test/subdir2")

        listing = await self.sandbox.fs.ls("/tmp/ls-subdir-test")

        assert listing.subdirectories is not None
        names = [d.name for d in listing.subdirectories]
        assert "subdir1" in names
        assert "subdir2" in names

    async def test_returns_file_metadata(self):
        """Test that file metadata is returned."""
        await self.sandbox.fs.write("/tmp/meta-test.txt", "some content")
        listing = await self.sandbox.fs.ls("/tmp")

        file = next((f for f in listing.files if f.name == "meta-test.txt"), None)
        assert file is not None
        assert file.path == "/tmp/meta-test.txt"


@pytest.mark.asyncio(loop_scope="class")
class TestMkdir(TestFilesystemOperations):
    """Test mkdir operations."""

    async def test_creates_directory(self):
        """Test creating a directory."""
        import time

        path = f"/tmp/new-dir-{int(time.time() * 1000)}"
        await self.sandbox.fs.mkdir(path)

        listing = await self.sandbox.fs.ls(path)
        assert listing is not None

    async def test_creates_nested_directories(self):
        """Test creating nested directories."""
        import time

        path = f"/tmp/nested-{int(time.time() * 1000)}/level1/level2"
        await self.sandbox.fs.mkdir(path)

        listing = await self.sandbox.fs.ls(path)
        assert listing is not None


@pytest.mark.asyncio(loop_scope="class")
class TestCopy(TestFilesystemOperations):
    """Test cp (copy) operations."""

    async def test_copies_file(self):
        """Test copying a file."""
        src = "/tmp/cp-src.txt"
        dst = "/tmp/cp-dst.txt"

        await self.sandbox.fs.write(src, "copy me")
        await self.sandbox.fs.cp(src, dst)

        content = await self.sandbox.fs.read(dst)
        assert content == "copy me"

    async def test_copies_directory(self):
        """Test copying a directory."""
        src_dir = "/tmp/cp-dir-src"
        dst_dir = "/tmp/cp-dir-dst"

        await self.sandbox.fs.write(f"{src_dir}/file.txt", "content")
        await self.sandbox.fs.cp(src_dir, dst_dir)

        content = await self.sandbox.fs.read(f"{dst_dir}/file.txt")
        assert content == "content"


@pytest.mark.asyncio(loop_scope="class")
class TestRemove(TestFilesystemOperations):
    """Test rm (remove) operations."""

    async def test_removes_file(self):
        """Test removing a file."""
        path = "/tmp/rm-file.txt"
        await self.sandbox.fs.write(path, "delete me")

        await self.sandbox.fs.rm(path)

        # File should no longer exist
        with pytest.raises(Exception):
            await self.sandbox.fs.read(path)

    async def test_removes_directory_recursively(self):
        """Test removing a directory recursively."""
        dir_path = "/tmp/rm-dir"
        await self.sandbox.fs.write(f"{dir_path}/file.txt", "content")
        await self.sandbox.fs.mkdir(f"{dir_path}/subdir")

        await self.sandbox.fs.rm(dir_path, recursive=True)

        with pytest.raises(Exception):
            await self.sandbox.fs.ls(dir_path)


@pytest.mark.asyncio(loop_scope="class")
class TestWatch(TestFilesystemOperations):
    """Test watch operations."""

    async def test_watches_for_file_changes(self):
        """Test watching for file changes."""
        import time

        dir_path = f"/tmp/watch-test-{int(time.time() * 1000)}"
        await self.sandbox.fs.mkdir(dir_path)

        change_detected = False

        def on_change(event):
            nonlocal change_detected
            if event.name == "watched-file.txt":
                change_detected = True

        handle = self.sandbox.fs.watch(dir_path, on_change)

        try:
            await async_sleep(0.5)
            # Trigger a file change
            await self.sandbox.fs.write(f"{dir_path}/watched-file.txt", "new content")

            # Wait for callback
            await async_sleep(0.5)
            assert change_detected is True
        finally:
            handle.close()


@pytest.mark.asyncio(loop_scope="class")
class TestMultipartUpload(TestFilesystemOperations):
    """Test multipart upload for large files."""

    async def test_uploads_small_file(self):
        """Test uploading a small file (< 1MB) via regular upload."""
        content = "Hello, world! " * 1000  # ~14KB
        path = "/tmp/small-upload.txt"

        await self.sandbox.fs.write(path, content)

        result = await self.sandbox.fs.read(path)
        assert result == content

    async def test_uploads_large_text_file(self):
        """Test uploading a large text file (> 1MB) via multipart."""
        content = "Large file content line. " * 50000  # ~1.2MB
        path = "/tmp/large-upload.txt"

        await self.sandbox.fs.write(path, content)

        result = await self.sandbox.fs.read(path)
        assert len(result) == len(content)
        assert result == content


@pytest.mark.asyncio(loop_scope="class")
class TestFind(TestFilesystemOperations):
    """Test find operations."""

    async def test_finds_files_by_pattern(self):
        """Test finding files by pattern."""
        import time

        base_dir = f"/tmp/find-test-{int(time.time() * 1000)}"
        await self.sandbox.fs.write(f"{base_dir}/file1.txt", "content1")
        await self.sandbox.fs.write(f"{base_dir}/file2.txt", "content2")
        await self.sandbox.fs.write(f"{base_dir}/file3.py", "content3")

        result = await self.sandbox.fs.find(base_dir, patterns=["*.txt"])
        assert result is not None

        assert result.total >= 2
        paths = [m.path for m in result.matches]
        assert any("file1.txt" in p for p in paths)
        assert any("file2.txt" in p for p in paths)
        assert not any("file3.py" in p for p in paths)

    async def test_finds_directories(self):
        """Test finding directories only."""
        import time

        base_dir = f"/tmp/find-dir-test-{int(time.time() * 1000)}"
        await self.sandbox.fs.mkdir(f"{base_dir}/subdir1")
        await self.sandbox.fs.mkdir(f"{base_dir}/subdir2")
        await self.sandbox.fs.write(f"{base_dir}/file.txt", "content")

        result = await self.sandbox.fs.find(base_dir, type="directory")
        assert result is not None

        types = [m.type_ for m in result.matches]
        assert all(t == "directory" for t in types)

    async def test_finds_files_only(self):
        """Test finding files only."""
        import time

        base_dir = f"/tmp/find-files-only-{int(time.time() * 1000)}"
        await self.sandbox.fs.mkdir(f"{base_dir}/subdir")
        await self.sandbox.fs.write(f"{base_dir}/file1.txt", "content1")
        await self.sandbox.fs.write(f"{base_dir}/file2.txt", "content2")

        result = await self.sandbox.fs.find(base_dir, type="file")
        assert result is not None

        assert result.matches is not None
        types = [m.type_ for m in result.matches]
        assert all(t == "file" for t in types)

    async def test_respects_max_results(self):
        """Test that max_results limits the results."""
        import time

        base_dir = f"/tmp/find-max-{int(time.time() * 1000)}"
        for i in range(5):
            await self.sandbox.fs.write(f"{base_dir}/file{i}.txt", f"content{i}")

        result = await self.sandbox.fs.find(base_dir, max_results=2)
        assert result is not None

        assert result.matches is not None
        assert len(result.matches) <= 2

    async def test_finds_nested_files(self):
        """Test finding files in nested directories."""
        import time

        base_dir = f"/tmp/find-nested-{int(time.time() * 1000)}"
        await self.sandbox.fs.write(f"{base_dir}/level1/level2/deep.txt", "deep content")
        await self.sandbox.fs.write(f"{base_dir}/shallow.txt", "shallow content")

        result = await self.sandbox.fs.find(base_dir, patterns=["*.txt"])
        assert result is not None

        assert result.matches is not None
        paths = [m.path for m in result.matches]
        assert any("deep.txt" in p for p in paths)
        assert any("shallow.txt" in p for p in paths)

    async def test_find_with_all_options(self):
        """Test find with all options set."""
        import time

        base_dir = f"/tmp/find-all-opts-{int(time.time() * 1000)}"
        # Create visible files
        await self.sandbox.fs.write(f"{base_dir}/visible1.txt", "content1")
        await self.sandbox.fs.write(f"{base_dir}/visible2.txt", "content2")
        await self.sandbox.fs.write(f"{base_dir}/visible3.py", "content3")
        # Create hidden file
        await self.sandbox.fs.write(f"{base_dir}/.hidden.txt", "hidden content")
        # Create file in excluded directory
        await self.sandbox.fs.write(f"{base_dir}/node_modules/dep.txt", "dep content")
        # Create nested file
        await self.sandbox.fs.write(f"{base_dir}/src/nested.txt", "nested content")

        result = await self.sandbox.fs.find(
            base_dir,
            type="file",
            patterns=["*.txt"],
            max_results=10,
            exclude_dirs=["node_modules"],
            exclude_hidden=True,
        )
        assert result is not None

        assert result.matches is not None
        paths = [m.path for m in result.matches]
        types = [m.type_ for m in result.matches]

        # Should find visible .txt files
        assert any("visible1.txt" in p for p in paths)
        assert any("visible2.txt" in p for p in paths)
        assert any("nested.txt" in p for p in paths)
        # Should not find .py file (pattern filter)
        assert not any("visible3.py" in p for p in paths)
        # Should not find hidden file (exclude_hidden)
        assert not any(".hidden.txt" in p for p in paths)
        # Should not find file in excluded directory
        assert not any("node_modules" in p for p in paths)
        # All should be files
        assert all(t == "file" for t in types)


@pytest.mark.asyncio(loop_scope="class")
class TestGrep(TestFilesystemOperations):
    """Test grep (content search) operations."""

    async def test_searches_text_in_files(self):
        """Test basic text search."""
        import time

        base_dir = f"/tmp/grep-test-{int(time.time() * 1000)}"
        await self.sandbox.fs.write(f"{base_dir}/file1.txt", "Hello World")
        await self.sandbox.fs.write(f"{base_dir}/file2.txt", "Goodbye World")
        await self.sandbox.fs.write(f"{base_dir}/file3.txt", "No match here")

        result = await self.sandbox.fs.grep("World", base_dir)
        assert result is not None

        assert result.matches is not None
        assert result.total >= 2
        assert result.query == "World"
        paths = [m.path for m in result.matches]
        assert any("file1.txt" in p for p in paths)
        assert any("file2.txt" in p for p in paths)

    async def test_case_sensitive_search(self):
        """Test case sensitive search."""
        import time

        base_dir = f"/tmp/grep-case-{int(time.time() * 1000)}"
        await self.sandbox.fs.write(f"{base_dir}/upper.txt", "HELLO")
        await self.sandbox.fs.write(f"{base_dir}/lower.txt", "hello")

        result = await self.sandbox.fs.grep("HELLO", base_dir, case_sensitive=True)
        assert result is not None

        assert result.matches is not None
        paths = [m.path for m in result.matches]
        assert any("upper.txt" in p for p in paths)
        # In case sensitive mode, lowercase should not match
        assert len([p for p in paths if "lower.txt" in p]) == 0

    async def test_case_insensitive_search(self):
        """Test case insensitive search."""
        import time

        base_dir = f"/tmp/grep-icase-{int(time.time() * 1000)}"
        await self.sandbox.fs.write(f"{base_dir}/upper.txt", "HELLO")
        await self.sandbox.fs.write(f"{base_dir}/lower.txt", "hello")

        result = await self.sandbox.fs.grep("hello", base_dir, case_sensitive=False)
        assert result is not None

        assert result.matches is not None
        paths = [m.path for m in result.matches]
        assert any("upper.txt" in p for p in paths)
        assert any("lower.txt" in p for p in paths)

    async def test_returns_line_information(self):
        """Test that search returns line and column info."""
        import time

        base_dir = f"/tmp/grep-line-{int(time.time() * 1000)}"
        content = "first line\nsecond line with match\nthird line"
        await self.sandbox.fs.write(f"{base_dir}/file.txt", content)

        result = await self.sandbox.fs.grep("match", base_dir)
        assert result is not None

        assert result.matches is not None
        assert len(result.matches) >= 1
        match = result.matches[0]
        assert match.line is not None
        assert match.text is not None
        assert "match" in match.text

    async def test_respects_file_pattern(self):
        """Test searching only in files matching pattern."""
        import time

        base_dir = f"/tmp/grep-pattern-{int(time.time() * 1000)}"
        await self.sandbox.fs.write(f"{base_dir}/code.py", "search_term here")
        await self.sandbox.fs.write(f"{base_dir}/data.txt", "search_term here too")

        result = await self.sandbox.fs.grep("search_term", base_dir, file_pattern="*.py")
        assert result is not None

        assert result.matches is not None
        paths = [m.path for m in result.matches]
        assert any("code.py" in p for p in paths)
        assert not any("data.txt" in p for p in paths)

    async def test_respects_max_results(self):
        """Test that max_results limits the search results."""
        import time

        base_dir = f"/tmp/grep-max-{int(time.time() * 1000)}"
        for i in range(5):
            await self.sandbox.fs.write(f"{base_dir}/file{i}.txt", "findme content")

        result = await self.sandbox.fs.grep("findme", base_dir, max_results=2)
        assert result is not None

        assert result.matches is not None
        assert len(result.matches) <= 2

    async def test_searches_nested_directories(self):
        """Test searching in nested directories."""
        import time

        base_dir = f"/tmp/grep-nested-{int(time.time() * 1000)}"
        await self.sandbox.fs.write(f"{base_dir}/level1/level2/deep.txt", "unique_string")
        await self.sandbox.fs.write(f"{base_dir}/shallow.txt", "unique_string")

        result = await self.sandbox.fs.grep("unique_string", base_dir)
        assert result is not None

        assert result.matches is not None
        assert result.total >= 2
        paths = [m.path for m in result.matches]
        assert any("deep.txt" in p for p in paths)
        assert any("shallow.txt" in p for p in paths)

    async def test_grep_with_all_options(self):
        """Test grep with all options set."""
        import time

        base_dir = f"/tmp/grep-all-opts-{int(time.time() * 1000)}"
        # Create files with target content
        content_with_context = "line one\nline two\nSEARCH_TARGET here\nline four\nline five"
        await self.sandbox.fs.write(f"{base_dir}/main.py", content_with_context)
        await self.sandbox.fs.write(f"{base_dir}/other.txt", "SEARCH_TARGET in txt")
        await self.sandbox.fs.write(f"{base_dir}/lowercase.py", "search_target lowercase")
        # Create file in excluded directory
        await self.sandbox.fs.write(f"{base_dir}/node_modules/dep.py", "SEARCH_TARGET in excluded")

        result = await self.sandbox.fs.grep(
            "SEARCH_TARGET",
            base_dir,
            case_sensitive=True,
            context_lines=2,
            max_results=10,
            file_pattern="*.py",
            exclude_dirs=["node_modules"],
        )
        assert result is not None

        assert result.matches is not None
        assert result.query == "SEARCH_TARGET"
        paths = [m.path for m in result.matches]

        # Should find .py files with exact case match
        assert any("main.py" in p for p in paths)
        # Should not find .txt file (file_pattern filter)
        assert not any("other.txt" in p for p in paths)
        # Should not find lowercase version (case_sensitive)
        assert not any("lowercase.py" in p for p in paths)
        # Should not find file in excluded directory
        assert not any("node_modules" in p for p in paths)

        # Check context lines are included for main.py match
        main_match = next((m for m in result.matches if "main.py" in m.path), None)
        assert main_match is not None
        # With context_lines=2, we should have context around the match
        if main_match.context:
            assert "line" in main_match.context or "SEARCH_TARGET" in main_match.context


@pytest.mark.asyncio(loop_scope="class")
class TestParallelOperations(TestFilesystemOperations):
    """Test parallel operations."""

    async def test_handles_100_parallel_file_reads(self):
        """Test handling 100 parallel file reads."""
        # Create a test file
        content = "A" * (200 * 1024)  # 200KB
        path = "/tmp/parallel-read.txt"
        await self.sandbox.fs.write(path, content)

        # Make 100 parallel read calls
        async def read_and_get_length():
            file_content = await self.sandbox.fs.read(path)
            return len(file_content)

        results = await asyncio.gather(*[read_and_get_length() for _ in range(100)])

        # All reads should return the same size
        assert all(size == len(content) for size in results)



"""Blaxel telemetry module."""

try:
    from .exporters import *  # noqa: F403, F401
    from .instrumentation import *  # noqa: F403, F401
    from .log import *  # noqa: F403, F401
except ImportError:
    pass

from .manager import *  # noqa: F403, F401



from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.expiration_policy_action import ExpirationPolicyAction
from ..models.expiration_policy_type import ExpirationPolicyType
from ..types import UNSET, Unset

T = TypeVar("T", bound="ExpirationPolicy")


@_attrs_define
class ExpirationPolicy:
    """Expiration policy for automatic sandbox cleanup based on time conditions

    Attributes:
        action (Union[Unset, ExpirationPolicyAction]): Action to take when the expiration condition is met Example:
            delete.
        type_ (Union[Unset, ExpirationPolicyType]): Type of expiration policy: ttl-idle (delete after inactivity), ttl-
            max-age (delete after total lifetime), or date (delete at specific time) Example: ttl-idle.
        value (Union[Unset, str]): Duration value for TTL policies (e.g., '30m', '24h', '7d') or ISO 8601 date for date
            policies Example: 24h.
    """

    action: Union[Unset, ExpirationPolicyAction] = UNSET
    type_: Union[Unset, ExpirationPolicyType] = UNSET
    value: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        action: Union[Unset, str] = UNSET
        if not isinstance(self.action, Unset):
            action = self.action.value

        type_: Union[Unset, str] = UNSET
        if not isinstance(self.type_, Unset):
            type_ = self.type_.value

        value = self.value

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if action is not UNSET:
            field_dict["action"] = action
        if type_ is not UNSET:
            field_dict["type"] = type_
        if value is not UNSET:
            field_dict["value"] = value

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T | None:
        if not src_dict:
            return None
        d = src_dict.copy()
        _action = d.pop("action", UNSET)
        action: Union[Unset, ExpirationPolicyAction]
        if isinstance(_action, Unset):
            action = UNSET
        else:
            action = ExpirationPolicyAction(_action)

        _type_ = d.pop("type", d.pop("type_", UNSET))
        type_: Union[Unset, ExpirationPolicyType]
        if isinstance(_type_, Unset):
            type_ = UNSET
        else:
            type_ = ExpirationPolicyType(_type_)

        value = d.pop("value", UNSET)

        expiration_policy = cls(
            action=action,
            type_=type_,
            value=value,
        )

        expiration_policy.additional_properties = d
        return expiration_policy

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.preview_token_metadata import PreviewTokenMetadata
    from ..models.preview_token_spec import PreviewTokenSpec


T = TypeVar("T", bound="PreviewToken")


@_attrs_define
class PreviewToken:
    """Token for a Preview

    Attributes:
        metadata (PreviewTokenMetadata): PreviewTokenMetadata
        spec (PreviewTokenSpec): Spec for a Preview Token
    """

    metadata: "PreviewTokenMetadata"
    spec: "PreviewTokenSpec"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        if type(self.metadata) is dict:
            metadata = self.metadata
        else:
            metadata = self.metadata.to_dict()

        if type(self.spec) is dict:
            spec = self.spec
        else:
            spec = self.spec.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "metadata": metadata,
                "spec": spec,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T | None:
        from ..models.preview_token_metadata import PreviewTokenMetadata
        from ..models.preview_token_spec import PreviewTokenSpec

        if not src_dict:
            return None
        d = src_dict.copy()
        metadata = PreviewTokenMetadata.from_dict(d.pop("metadata"))

        spec = PreviewTokenSpec.from_dict(d.pop("spec"))

        preview_token = cls(
            metadata=metadata,
            spec=spec,
        )

        preview_token.additional_properties = d
        return preview_token

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Google ADK Tools Integration Tests."""

pytest_plugins = []
import pytest  # noqa: E402

pytest.importorskip(
    "google.adk", reason="google-adk not installed (install with: blaxel[googleadk])"
)

import pytest_asyncio  # noqa: E402
from google.adk.agents import Agent  # noqa: E402
from google.adk.runners import Runner  # noqa: E402
from google.adk.sessions import InMemorySessionService  # noqa: E402
from google.genai import types  # noqa: E402

from blaxel.core.sandbox import SandboxInstance  # noqa: E402
from blaxel.googleadk import bl_model, bl_tools  # noqa: E402
from tests.helpers import default_image, default_labels, unique_name  # noqa: E402


@pytest.mark.asyncio(loop_scope="class")
class TestBlTools:
    """Test bl_tools functionality."""

    sandbox: SandboxInstance | None = None
    sandbox_name: str | None = None

    @pytest_asyncio.fixture(autouse=True, scope="class", loop_scope="class")
    async def setup_sandbox(self, request):
        """Set up a sandbox for the test class."""
        request.cls.sandbox_name = unique_name("googleadk-tools-test")
        request.cls.sandbox = await SandboxInstance.create(
            {
                "name": request.cls.sandbox_name,
                "image": default_image,
                "memory": 2048,
                "labels": default_labels,
            }
        )

        yield

        # Cleanup
        try:
            await request.cls.sandbox.delete()
        except Exception:
            pass

    async def test_agent_can_use_tools(self):
        """Test that an agent can use sandbox tools to list files."""
        model = await bl_model("sandbox-openai")
        tools = await bl_tools([f"sandbox/{self.sandbox_name}"])

        agent = Agent(
            name="test",
            model=model,
            tools=tools,
            instruction="You are a helpful assistant. Use the tools available to answer the user's question.",
        )

        session_service = InMemorySessionService()
        session = await session_service.create_session(app_name="test", user_id="test-user")

        runner = Runner(agent=agent, app_name="test", session_service=session_service)

        message = types.Content(
            parts=[types.Part(text="List the files and directories in /")],
            role="user",
        )

        events = []
        async for event in runner.run_async(
            user_id="test-user",
            session_id=session.id,
            new_message=message,
        ):
            events.append(event)

        assert len(events) > 0
        # Check that at least one event has content
        has_content = any(event.content and event.content.parts for event in events)
        assert has_content
        # Print the last event with content
        for event in reversed(events):
            if event.content and event.content.parts:
                text = " ".join(p.text for p in event.content.parts if p.text)
                print(f"\n[googleadk agent result] {text}")
                break



from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.env import Env
    from ..models.port import Port


T = TypeVar("T", bound="SandboxRuntime")


@_attrs_define
class SandboxRuntime:
    """Runtime configuration defining how the sandbox VM is provisioned and its resource limits

    Attributes:
        envs (Union[Unset, list['Env']]): Environment variables injected into the sandbox. Supports Kubernetes EnvVar
            format with valueFrom references.
        expires (Union[Unset, str]): Absolute expiration timestamp in ISO 8601 format when the sandbox will be deleted
            Example: 2025-12-31T23:59:59Z.
        image (Union[Unset, str]): Sandbox image to use. Can be a public Blaxel image (e.g., blaxel/base-image:latest)
            or a custom template image built with 'bl deploy'. Example: blaxel/base-image:latest.
        memory (Union[Unset, int]): Memory allocation in megabytes. Also determines CPU allocation (CPU cores = memory
            in MB / 2048, e.g., 4096MB = 2 CPUs). Example: 4096.
        ports (Union[Unset, list['Port']]): Set of ports for a resource
        ttl (Union[Unset, str]): Time-to-live duration after which the sandbox is automatically deleted (e.g., '30m',
            '24h', '7d') Example: 24h.
    """

    envs: Union[Unset, list["Env"]] = UNSET
    expires: Union[Unset, str] = UNSET
    image: Union[Unset, str] = UNSET
    memory: Union[Unset, int] = UNSET
    ports: Union[Unset, list["Port"]] = UNSET
    ttl: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        envs: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.envs, Unset):
            envs = []
            for envs_item_data in self.envs:
                if type(envs_item_data) is dict:
                    envs_item = envs_item_data
                else:
                    envs_item = envs_item_data.to_dict()
                envs.append(envs_item)

        expires = self.expires

        image = self.image

        memory = self.memory

        ports: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.ports, Unset):
            ports = []
            for componentsschemas_ports_item_data in self.ports:
                if type(componentsschemas_ports_item_data) is dict:
                    componentsschemas_ports_item = componentsschemas_ports_item_data
                else:
                    componentsschemas_ports_item = componentsschemas_ports_item_data.to_dict()
                ports.append(componentsschemas_ports_item)

        ttl = self.ttl

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if envs is not UNSET:
            field_dict["envs"] = envs
        if expires is not UNSET:
            field_dict["expires"] = expires
        if image is not UNSET:
            field_dict["image"] = image
        if memory is not UNSET:
            field_dict["memory"] = memory
        if ports is not UNSET:
            field_dict["ports"] = ports
        if ttl is not UNSET:
            field_dict["ttl"] = ttl

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T | None:
        from ..models.env import Env
        from ..models.port import Port

        if not src_dict:
            return None
        d = src_dict.copy()
        envs = []
        _envs = d.pop("envs", UNSET)
        for envs_item_data in _envs or []:
            envs_item = Env.from_dict(envs_item_data)

            envs.append(envs_item)

        expires = d.pop("expires", UNSET)

        image = d.pop("image", UNSET)

        memory = d.pop("memory", UNSET)

        ports = []
        _ports = d.pop("ports", UNSET)
        for componentsschemas_ports_item_data in _ports or []:
            componentsschemas_ports_item = Port.from_dict(componentsschemas_ports_item_data)

            ports.append(componentsschemas_ports_item)

        ttl = d.pop("ttl", UNSET)

        sandbox_runtime = cls(
            envs=envs,
            expires=expires,
            image=image,
            memory=memory,
            ports=ports,
            ttl=ttl,
        )

        sandbox_runtime.additional_properties = d
        return sandbox_runtime

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
